import asyncio
import contextlib
from dataclasses import replace
import logging
import pathlib
from collections import OrderedDict
from typing import (Any, Dict, Union, List, Optional, Tuple,
                    TYPE_CHECKING, cast, overload, Sequence)

from opentrons_shared_data.pipette import name_config
from opentrons import types as top_types
from opentrons.util import linal
from functools import lru_cache
from opentrons.config import robot_configs
from opentrons.config.types import RobotConfig

from .util import (
    use_or_initialize_loop, DeckTransformState, check_motion_bounds)
from .pipette import (
    Pipette, generate_hardware_configs, load_from_config_and_check_skip)
from .controller import Controller
from .simulator import Simulator
from .constants import (SHAKE_OFF_TIPS_SPEED, SHAKE_OFF_TIPS_DROP_DISTANCE,
                        SHAKE_OFF_TIPS_PICKUP_DISTANCE,
                        DROP_TIP_RELEASE_DISTANCE)
from .execution_manager import ExecutionManager
from .pause_manager import PauseManager
from .module_control import AttachedModulesControl
from .types import (Axis, HardwareAPILike, CriticalPoint,
                    MustHomeError, NoTipAttachedError, DoorState,
                    DoorStateNotification, PipettePair, TipAttachedError,
                    HardwareAction, PairedPipetteConfigValueError,
                    MotionChecks, PauseType)
from . import modules, robot_calibration as rb_cal

if TYPE_CHECKING:
    from opentrons_shared_data.pipette.dev_types import (
        UlPerMmAction, PipetteName
    )
    from .dev_types import PipetteDict


mod_log = logging.getLogger(__name__)


InstrumentsByMount = Dict[top_types.Mount, Optional[Pipette]]
PipetteHandlingData = Tuple[Pipette, top_types.Mount]


class API(HardwareAPILike):
    """ This API is the primary interface to the hardware controller.

    Because the hardware manager controls access to the system's hardware
    as a whole, it is designed as a class of which only one should be
    instantiated at a time. This class's methods should be the only method
    of external access to the hardware. Each method may be minimal - it may
    only delegate the call to another submodule of the hardware manager -
    but its purpose is to be gathered here to provide a single interface.
    """

    CLS_LOG = mod_log.getChild('API')

    def __init__(self,
                 backend: Union[Controller, Simulator],
                 loop: asyncio.AbstractEventLoop,
                 config: RobotConfig
                 ) -> None:
        """ Initialize an API instance.

        This should rarely be explicitly invoked by an external user; instead,
        one of the factory methods build_hardware_controller or
        build_hardware_simulator should be used.
        """
        self._log = self.CLS_LOG.getChild(str(id(self)))
        self._config = config
        self._backend = backend
        self._loop = loop

        self._execution_manager = ExecutionManager(loop=loop)
        self._callbacks: set = set()
        # {'X': 0.0, 'Y': 0.0, 'Z': 0.0, 'A': 0.0, 'B': 0.0, 'C': 0.0}
        self._current_position: Dict[Axis, float] = {}

        self._attached_instruments: InstrumentsByMount = {
            top_types.Mount.LEFT: None,
            top_types.Mount.RIGHT: None
        }
        self._last_moved_mount: Optional[top_types.Mount] = None
        # The motion lock synchronizes calls to long-running physical tasks
        # involved in motion. This fixes issue where for instance a move()
        # or home() call is in flight and something else calls
        # current_position(), which will not be updated until the move() or
        # home() call succeeds or fails.
        self._motion_lock = asyncio.Lock(loop=self._loop)
        self._door_state = DoorState.CLOSED
        self._robot_calibration = rb_cal.load()
        self._pause_manager = PauseManager(self._door_state)

    @property
    def robot_calibration(self) -> rb_cal.RobotCalibration:
        return self._robot_calibration

    def reset_robot_calibration(self):
        self._calculate_valid_attitude.cache_clear()
        self._robot_calibration = rb_cal.load()

    def set_robot_calibration(
            self, robot_calibration: rb_cal.RobotCalibration):
        self._calculate_valid_attitude.cache_clear()
        self._robot_calibration = robot_calibration

    @property
    def door_state(self) -> DoorState:
        return self._door_state

    @door_state.setter
    def door_state(self, door_state: DoorState):
        self._door_state = door_state

    def _update_door_state(self, door_state: DoorState):
        mod_log.info(
            f'Updating the window switch status: {door_state}')
        self.door_state = door_state
        self._pause_manager.set_door(self.door_state)
        for cb in self._callbacks:
            hw_event = DoorStateNotification(
                new_state=door_state,
                blocking=self._pause_manager.blocked_by_door)
            try:
                cb(hw_event)
            except Exception:
                mod_log.exception('Errored during door state event callback')

    def _reset_last_mount(self):
        self._last_moved_mount = None

    @classmethod
    async def build_hardware_controller(
            cls, config: RobotConfig = None,
            port: str = None,
            loop: asyncio.AbstractEventLoop = None,
            firmware: Tuple[pathlib.Path, str] = None) -> 'API':
        """ Build a hardware controller that will actually talk to hardware.

        This method should not be used outside of a real robot, and on a
        real robot only one true hardware controller may be active at one
        time.

        :param config: A config to preload. If not specified, load the default.
        :param port: A port to connect to. If not specified, the default port
                     (found by scanning for connected FT232Rs).
        :param loop: An event loop to use. If not specified, use the result of
                     :py:meth:`asyncio.get_event_loop`.
        """
        checked_loop = use_or_initialize_loop(loop)
        checked_config = config or robot_configs.load()
        backend = await Controller.build(checked_config)
        backend.set_lights(button=None, rails=False)

        async def blink():
            while True:
                backend.set_lights(button=True, rails=None)
                await asyncio.sleep(0.5)
                backend.set_lights(button=False, rails=None)
                await asyncio.sleep(0.5)

        blink_task = checked_loop.create_task(blink())
        try:
            try:
                await backend.connect(port)
                fw_version = backend.fw_version
            except Exception:
                mod_log.exception(
                    'Motor driver could not connect, reprogramming if possible'
                )
                fw_version = None

            if firmware is not None:
                if fw_version != firmware[1]:
                    await backend.update_firmware(
                        str(firmware[0]), checked_loop, True)
                    await backend.connect(port)
            elif firmware is None and fw_version is None:
                msg = 'Motor controller could not be connected and no '\
                    'firmware was provided for (re)programming'
                mod_log.error(msg)
                raise RuntimeError(msg)

            api_instance = cls(backend, loop=checked_loop, config=checked_config)
            await api_instance.cache_instruments()
            module_controls = await AttachedModulesControl.build(api_instance)
            backend.module_controls = module_controls
            checked_loop.create_task(backend.watch(loop=checked_loop))
            backend.start_gpio_door_watcher(
                loop=checked_loop,
                update_door_state=api_instance._update_door_state)
            return api_instance
        finally:
            blink_task.cancel()

    @classmethod
    async def build_hardware_simulator(
            cls,
            attached_instruments: Dict[top_types.Mount, Dict[str, Optional[str]]] = None,  # noqa: E501
            attached_modules: List[str] = None,
            config: RobotConfig = None,
            loop: asyncio.AbstractEventLoop = None,
            strict_attached_instruments: bool = True) -> 'API':
        """ Build a simulating hardware controller.

        This method may be used both on a real robot and on dev machines.
        Multiple simulating hardware controllers may be active at one time.
        """

        if None is attached_instruments:
            attached_instruments = {}

        if None is attached_modules:
            attached_modules = []

        checked_loop = use_or_initialize_loop(loop)
        checked_config = config or robot_configs.load()
        backend = await Simulator.build(
            attached_instruments,
            attached_modules,
            checked_config, checked_loop,
            strict_attached_instruments)
        api_instance = cls(backend, loop=checked_loop, config=checked_config)
        await api_instance.cache_instruments()
        module_controls = await AttachedModulesControl.build(api_instance)
        backend.module_controls = module_controls
        await backend.watch()
        return api_instance

    def __repr__(self):
        return '<{} using backend {}>'.format(type(self),
                                              type(self._backend))

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        """ The event loop used by this instance. """
        return self._loop

    def set_loop(self, loop: asyncio.AbstractEventLoop):
        self._loop = loop
        self._motion_lock = asyncio.Lock(loop=loop)

    @property
    def is_simulator(self):
        """ `True` if this is a simulator; `False` otherwise. """
        return isinstance(self._backend, Simulator)

    def validate_calibration(self) -> DeckTransformState:
        """
        The lru cache decorator is currently not supported by the
        ThreadManager. To work around this, we need to wrap the
        actualy function around a dummy outer function.

        Once decorators are more fully supported, we can remove this.
        """
        return self._calculate_valid_attitude()

    @lru_cache(maxsize=1)
    def _calculate_valid_attitude(self) -> DeckTransformState:
        return rb_cal.validate_attitude_deck_calibration(
            self._robot_calibration.deck_calibration)

    async def register_callback(self, cb):
        """ Allows the caller to register a callback, and returns a closure
        that can be used to unregister the provided callback
        """
        self._callbacks.add(cb)

        def unregister():
            self._callbacks.remove(cb)

        return unregister

    def get_fw_version(self) -> str:
        """
        Return the firmware version of the connected hardware.

        The version is a string retrieved directly from the attached hardware
        (or possibly simulator).
        """
        from_backend = self._backend.fw_version
        if from_backend is None:
            return 'unknown'
        else:
            return from_backend

    @property
    def fw_version(self) -> Optional[str]:
        return self.get_fw_version()

    @property
    def board_revision(self) -> str:
        return str(self._backend.board_revision)

    # Incidentals (i.e. not motion) API

    async def set_lights(self, button: bool = None, rails: bool = None):
        """ Control the robot lights.

        :param button: If specified, turn the button light on (`True`) or
                       off (`False`). If not specified, do not change the
                       button light.
        :param rails: If specified, turn the rail lights on (`True`) or
                      off (`False`). If not specified, do not change the
                      rail lights.
        """
        self._backend.set_lights(button, rails)

    def get_lights(self) -> Dict[str, bool]:
        """ Return the current status of the robot lights.

        :returns: A dict of the lights: `{'button': bool, 'rails': bool}`
        """
        return self._backend.get_lights()

    async def identify(self, duration_s: int = 5):
        """ Blink the button light to identify the robot.

        :param int duration_s: The duration to blink for, in seconds.
        """
        count = duration_s * 4
        on = False
        for sec in range(count):
            then = self._loop.time()
            await self.set_lights(button=on)
            on = not on
            now = self._loop.time()
            await asyncio.sleep(max(0, 0.25 - (now - then)))
        await self.set_lights(button=True)

    async def delay(self, duration_s: float):
        """ Delay execution by pausing and sleeping.
        """
        await self._wait_for_is_running()
        self.pause(PauseType.DELAY)
        try:
            if not self.is_simulator:
                async def sleep_for_seconds(seconds: float):
                    await asyncio.sleep(seconds)
                delay_task = self._loop.create_task(sleep_for_seconds(duration_s))
                await self._execution_manager.register_cancellable_task(delay_task)
        finally:
            self.resume(PauseType.DELAY)

    def reset_instrument(self, mount: top_types.Mount = None):
        """
        Reset the internal state of a pipette by its mount, without doing
        any lower level reconfiguration. This is useful to make sure that no
        settings changes from a protocol persist.

        :param mount: If specified, reset that mount. If not specified,
                      reset both
        """
        def _reset(m: top_types.Mount):
            self._log.info(f"Resetting configuration for {m}")
            p = self._attached_instruments[m]
            if not p:
                return
            new_p = Pipette(
                p._config,
                rb_cal.load_pipette_offset(p.pipette_id, m),
                p.pipette_id)
            new_p.act_as(p.acting_as)
            self._attached_instruments[m] = new_p

        if not mount:
            for m in top_types.Mount:
                _reset(m)
        else:
            _reset(mount)

    async def cache_instruments(
            self,
            require: Dict[top_types.Mount, 'PipetteName'] = None):
        """
        Scan the attached instruments, take necessary configuration actions,
        and set up hardware controller internal state if necessary.

        :param require: If specified, the require should be a dict
                        of mounts to instrument names describing
                        the instruments expected to be present. This can
                        save a subsequent of :py:attr:`attached_instruments`
                        and also serves as the hook for the hardware
                        simulator to decide what is attached.
        :raises RuntimeError: If an instrument is expected but not found.

        .. note::

            This function will only change the things that need to be changed.
            If the same pipette (by serial) or the same lack of pipette is
            observed on a mount before and after the scan, no action will be
            taken. That makes this function appropriate for setting up the
            robot for operation, but not for making sure that any previous
            settings changes have been reset. For the latter use case, use
            :py:meth:`reset_instrument`.

        """
        self._log.info("Updating instrument model cache")
        checked_require = require or {}
        for mount, name in checked_require.items():
            if name not in name_config():
                raise RuntimeError(f'{name} is not a valid pipette name')
        found = self._backend.get_attached_instruments(checked_require)

        for mount, instrument_data in found.items():
            config = instrument_data.get('config')
            req_instr = checked_require.get(mount, None)
            pip_id = instrument_data.get('id')
            pip_offset_cal = rb_cal.load_pipette_offset(pip_id, mount)
            p, may_skip = load_from_config_and_check_skip(
                config,
                self._attached_instruments[mount],
                req_instr,
                pip_id,
                pip_offset_cal)
            self._attached_instruments[mount] = p
            if req_instr and p:
                p.act_as(req_instr)

            if may_skip:
                self._log.info(
                    f"Skipping configuration on {mount.name}")
                continue

            self._log.info(
                f"Doing full configuration on {mount.name}")
            hw_config = generate_hardware_configs(
                p, self._config, self._backend.board_revision)
            self._backend.configure_mount(mount, hw_config)
        mod_log.info("Instruments found: {}".format(
            self._attached_instruments))

    def get_attached_instruments(self) -> Dict[top_types.Mount,
                                               'PipetteDict']:
        """ Get the status dicts of the cached attached instruments.

        Also available as :py:meth:`get_attached_instruments`.

        This returns a dictified version of the
        :py:class:`hardware_control.pipette.Pipette` as a dict keyed by
        the :py:class:`top_types.Mount` to which the pipette is attached.
        If no pipette is attached on a given mount, the mount key will
        still be present but will have the value ``None``.

        Note that this is only a query of a cached value; to actively scan
        for changes, use :py:meth:`cache_instruments`. This process deactivates
        the motors and should be used sparingly.
        """
        return {
            m: self.get_attached_instrument(m)
            for m in (top_types.Mount.LEFT, top_types.Mount.RIGHT)
        }

    def get_attached_instrument(self, mount: top_types.Mount) -> 'PipetteDict':
        instr = self._attached_instruments[mount]
        result: Dict[str, Any] = {}
        if instr:
            configs = ['name', 'min_volume', 'max_volume', 'channels',
                       'aspirate_flow_rate', 'dispense_flow_rate',
                       'pipette_id', 'current_volume', 'display_name',
                       'tip_length', 'model', 'blow_out_flow_rate',
                       'working_volume', 'tip_overlap', 'available_volume',
                       'return_tip_height', 'default_aspirate_flow_rates',
                       'default_blow_out_flow_rates',
                       'default_dispense_flow_rates']

            instr_dict = instr.as_dict()
            for key in configs:
                result[key] = instr_dict[key]
            result['has_tip'] = instr.has_tip
            result['aspirate_speed'] = self._plunger_speed(
                instr, instr.aspirate_flow_rate, 'aspirate')
            result['dispense_speed'] = self._plunger_speed(
                instr, instr.dispense_flow_rate, 'dispense')
            result['blow_out_speed'] = self._plunger_speed(
                instr, instr.blow_out_flow_rate, 'dispense')
            result['ready_to_aspirate'] = instr.ready_to_aspirate
            result['default_blow_out_speeds'] = {
                alvl: self._plunger_speed(instr, fr, 'dispense')
                for alvl, fr
                in instr.config.default_aspirate_flow_rates.items()
            }
            result['default_dispense_speeds'] = {
                alvl: self._plunger_speed(instr, fr, 'dispense')
                for alvl, fr
                in instr.config.default_dispense_flow_rates.items()
            }
            result['default_aspirate_speeds'] = {
                alvl: self._plunger_speed(instr, fr, 'aspirate')
                for alvl, fr
                in instr.config.default_aspirate_flow_rates.items()
            }
        return cast('PipetteDict', result)

    @property
    def attached_instruments(self) -> Dict[top_types.Mount, 'PipetteDict']:
        return self.get_attached_instruments()

    @property
    def attached_modules(self) -> List[modules.AbstractModule]:
        return self._backend.module_controls.available_modules

    async def update_firmware(
            self,
            firmware_file: str,
            loop: asyncio.AbstractEventLoop = None,
            explicit_modeset: bool = True) -> str:
        """ Update the firmware on the Smoothie board.

        :param firmware_file: The path to the firmware file.
        :param explicit_modeset: `True` to force the smoothie into programming
                                 mode; `False` to assume it is already in
                                 programming mode.
        :param loop: An asyncio event loop to use; if not specified, the one
                     associated with this instance will be used.
        :returns: The stdout of the tool used to update the smoothie
        """
        if None is loop:
            checked_loop = self._loop
        else:
            checked_loop = loop
        return await self._backend.update_firmware(firmware_file,
                                                   checked_loop,
                                                   explicit_modeset)

    # Global actions API
    def pause(self, pause_type: PauseType):
        """
        Pause motion of the robot after a current motion concludes.

        Individual calls to :py:meth:`move`
        (which :py:meth:`aspirate` and :py:meth:`dispense` and other
        calls may depend on) are considered atomic and will always complete if
        they have been called prior to a call to this method. However,
        subsequent calls to :py:meth:`move` that occur when the system
        is paused will not proceed until the system is resumed with
        :py:meth:`resume`.
        """
        self._pause_manager.pause(pause_type)

        async def _chained_calls():
            await self._execution_manager.pause()
            self._backend.pause()

        asyncio.run_coroutine_threadsafe(_chained_calls(), self._loop)

    def pause_with_message(self, message):
        self._log.warning('Pause with message: {}'.format(message))
        for cb in self._callbacks:
            cb(message)
        self.pause(PauseType.PAUSE)

    def resume(self, pause_type: PauseType):
        """
        Resume motion after a call to :py:meth:`pause`.
        """
        self._pause_manager.resume(pause_type)

        if self._pause_manager.should_pause:
            return

        # Resume must be called immediately to awaken thread running hardware
        #  methods (ThreadManager)
        self._backend.resume()

        async def _chained_calls():
            # mirror what happens API.pause.
            await self._execution_manager.resume()
            self._backend.resume()

        asyncio.run_coroutine_threadsafe(_chained_calls(), self._loop)

    def halt(self):
        """ Immediately stop motion.

        Calls to :py:meth:`stop` through the synch adapter while other calls
        are ongoing will typically wait until those calls are done, since most
        of the async calls here in fact block the loop while they talk to
        smoothie. To provide actual immediate halting, call this method which
        does not require use of the loop.

        After this call, the smoothie will be in a bad state until a call to
        :py:meth:`stop`.
        """
        self._backend.hard_halt()
        asyncio.run_coroutine_threadsafe(self._execution_manager.cancel(),
                                         self._loop)

    async def stop(self):
        """
        Stop motion as soon as possible, reset, and home.

        This will cancel motion (after the current call to :py:meth:`move`;
        see :py:meth:`pause` for more detail), then home and reset the
        robot.
        """
        self._backend.halt()
        self._log.info("Recovering from halt")
        await self.reset()
        await self.home()

    async def _wait_for_is_running(self):
        if not self.is_simulator:
            await self._execution_manager.wait_for_is_running()

    async def reset(self):
        """ Reset the stored state of the system.

        This will re-scan instruments and models, clearing any cached
        information about their presence or state.
        """
        self._pause_manager.reset()
        await self._execution_manager.reset()
        self._attached_instruments = {
            k: None for k in self._attached_instruments.keys()}
        await self.cache_instruments()

    # Gantry/frame (i.e. not pipette) action API
    async def home_z(self, mount: top_types.Mount = None):
        """ Home the two z-axes """
        self._reset_last_mount()
        if not mount:
            axes = [Axis.Z, Axis.A]
        else:
            axes = [Axis.by_mount(mount)]
        await self.home(axes)

    async def _do_plunger_home(
            self,
            axis: Axis = None,
            mount: top_types.Mount = None,
            acquire_lock: bool = True):
        assert (axis is not None) ^ (mount is not None),\
            'specify either axis or mount'
        if axis:
            checked_axis = axis
            checked_mount = Axis.to_mount(checked_axis)
        if mount:
            checked_mount = mount
            checked_axis = Axis.of_plunger(checked_mount)
        instr = self._attached_instruments[checked_mount]
        if not instr:
            return
        async with contextlib.AsyncExitStack() as stack:
            if acquire_lock:
                await stack.enter_async_context(self._motion_lock)
            with self._backend.save_current():
                self._backend.set_active_current(
                    {checked_axis: instr.config.plunger_current})
                self._backend.home([checked_axis.name.upper()])
                # either we were passed False for our acquire_lock and we
                # should pass it on, or we acquired the lock above and
                # shouldn't do it again
                await self._move_plunger(checked_mount,
                                         (instr.config.bottom,),
                                         acquire_lock=False)

    async def home_plunger(self, mount: top_types.Mount):
        """
        Home the plunger motor for a mount, and then return it to the 'bottom'
        position.

        :param mount: the mount associated with the target plunger
        :type mount: :py:class:`.top_types.Mount`
        """
        await self.current_position(mount=mount, refresh=True)
        await self._do_plunger_home(mount=mount,
                                    acquire_lock=True)

    async def home(self, axes: List[Axis] = None):
        """ Home the entire robot and initialize current position.
        :param axes: A list of axes to home. Default is `None`, which will
                     home everything.
        """
        await self._wait_for_is_running()
        self._reset_last_mount()
        # Initialize/update current_position
        checked_axes = axes or [ax for ax in Axis]
        gantry = [ax for ax in checked_axes if ax in Axis.gantry_axes()]
        smoothie_gantry = [ax.name.upper() for ax in gantry]
        smoothie_pos = {}
        plungers = [ax for ax in checked_axes
                    if ax not in Axis.gantry_axes()]

        async with self._motion_lock:
            if smoothie_gantry:
                smoothie_pos.update(self._backend.home(smoothie_gantry))
                self._current_position = self._deck_from_smoothie(smoothie_pos)
            for plunger in plungers:
                await self._do_plunger_home(axis=plunger, acquire_lock=False)

    async def add_tip(
            self,
            mount: top_types.Mount,
            tip_length: float):
        instr = self._attached_instruments[mount]
        attached = self.attached_instruments
        instr_dict = attached[mount]
        if instr and not instr.has_tip:
            instr.add_tip(tip_length=tip_length)
            instr_dict['has_tip'] = True
            instr_dict['tip_length'] = tip_length
        else:
            mod_log.warning('attach tip called while tip already attached')

    async def remove_tip(self, mount: top_types.Mount):
        instr = self._attached_instruments[mount]
        attached = self.attached_instruments
        instr_dict = attached[mount]
        if instr and instr.has_tip:
            instr.remove_tip()
            instr_dict['has_tip'] = False
            instr_dict['tip_length'] = 0.0
        else:
            mod_log.warning('detach tip called with no tip')

    def _deck_from_smoothie(
            self, smoothie_pos: Dict[str, float]) -> Dict[Axis, float]:
        """ Build a deck-abs position store from the smoothie's position

        This should take the smoothie style position {'X': float, etc}
        and turn it into the position dict used here {Axis.X: float} in
        deck-absolute coordinates. It runs the reverse deck transformation
        for the axes that require it.

        One piece of complexity is that if the gantry transformation includes
        a transition between non parallel planes, the z position of the left
        mount would depend on its actual position in deck frame, so we have
        to apply the mount offset.

        TODO: Figure out which frame the mount offset is measured in, because
              if it's measured in the deck frame (e.g. by touching off points
              on the deck) it has to go through the reverse transform to be
              added to the smoothie coordinates here.
        """
        with_enum = {Axis[k]: v for k, v in smoothie_pos.items()}
        plunger_axes = {k: v for k, v in with_enum.items()
                        if k not in Axis.gantry_axes()}
        right = (
            with_enum[Axis.X], with_enum[Axis.Y],
            with_enum[Axis.by_mount(top_types.Mount.RIGHT)])
        left = (with_enum[Axis.X],
                with_enum[Axis.Y],
                with_enum[Axis.by_mount(top_types.Mount.LEFT)])

        gantry_calibration =\
            self.robot_calibration.deck_calibration.attitude
        right_deck = linal.apply_reverse(
            gantry_calibration, right)
        left_deck = linal.apply_reverse(
            gantry_calibration, left)
        deck_pos = {Axis.X: right_deck[0],
                    Axis.Y: right_deck[1],
                    Axis.by_mount(top_types.Mount.RIGHT): right_deck[2],
                    Axis.by_mount(top_types.Mount.LEFT): left_deck[2]}
        deck_pos.update(plunger_axes)
        return deck_pos

    async def current_position(
            self,
            mount: top_types.Mount,
            critical_point: CriticalPoint = None,
            refresh: bool = False) -> Dict[Axis, float]:
        """ Return the postion (in deck coords) of the critical point of the
        specified mount.

        This returns cached position to avoid hitting the smoothie driver
        unless ``refresh`` is ``True``.

        If `critical_point` is specified, that critical point will be applied
        instead of the default one. For instance, if
        `critical_point=CriticalPoints.MOUNT` then the position of the mount
        will be returned. If the critical point specified does not exist, then
        the next one down is returned - for instance, if there is no tip on the
        specified mount but `CriticalPoint.TIP` was specified, the position of
        the nozzle will be returned.
        """
        if not self._current_position and not refresh:
            raise MustHomeError
        async with self._motion_lock:
            if refresh:
                self._current_position = self._deck_from_smoothie(
                    self._backend.update_position())
            if mount == top_types.Mount.RIGHT:
                offset = top_types.Point(0, 0, 0)
            else:
                offset = top_types.Point(*self._config.left_mount_offset)
            z_ax = Axis.by_mount(mount)
            plunger_ax = Axis.of_plunger(mount)
            cp = self._critical_point_for(mount, critical_point)
            return {
                Axis.X: self._current_position[Axis.X] + offset[0] + cp.x,
                Axis.Y: self._current_position[Axis.Y] + offset[1] + cp.y,
                z_ax: self._current_position[z_ax] + offset[2] + cp.z,
                plunger_ax: self._current_position[plunger_ax]
            }

    async def gantry_position(
            self,
            mount: top_types.Mount,
            critical_point: CriticalPoint = None,
            refresh: bool = False) -> top_types.Point:
        """ Return the position of the critical point as pertains to the gantry

        This ignores the plunger position and gives the Z-axis a predictable
        name (as :py:attr:`.Point.z`).

        `critical_point` specifies an override to the current critical point to
        use (see :py:meth:`current_position`).

        `refresh` if set to True, update the cached position using the
        smoothie driver (see :py:meth:`current_position`).
        """
        cur_pos = await self.current_position(mount, critical_point, refresh)
        return top_types.Point(x=cur_pos[Axis.X],
                               y=cur_pos[Axis.Y],
                               z=cur_pos[Axis.by_mount(mount)])

    @overload
    def _mounts(
        self,
        z_axis: PipettePair) -> Tuple[top_types.Mount, top_types.Mount]: ...

    @overload
    def _mounts(self, z_axis: top_types.Mount) -> Tuple[top_types.Mount]: ...

    def _mounts(self, z_axis):
        if isinstance(z_axis, PipettePair):
            return (z_axis.primary, z_axis.secondary)
        return (z_axis, )

    async def move_to(
            self, mount: Union[top_types.Mount, PipettePair],
            abs_position: top_types.Point,
            speed: float = None,
            critical_point: CriticalPoint = None,
            max_speeds: Dict[Axis, float] = None):
        """ Move the critical point of the specified mount to a location
        relative to the deck, at the specified speed. 'speed' sets the speed
        of all robot axes to the given value. So, if multiple axes are to be
        moved, they will do so at the same speed

        The critical point of the mount depends on the current status of
        the mount:
        - If the mount does not have anything attached, its critical point is
          the bottom of the mount attach bracket.
        - If the mount has a pipette attached and it is not known to have a
          pipette tip, the critical point is the end of the nozzle of a single
          pipette or the end of the backmost nozzle of a multipipette
        - If the mount has a pipette attached and it is known to have a
          pipette tip, the critical point is the end of the pipette tip for
          a single pipette or the end of the tip of the backmost nozzle of a
          multipipette

        :param mount: The mount to move
        :param abs_position: The target absolute position in
                             :ref:`protocol-api-deck-coords` to move the
                             critical point to
        :param speed: An overall head speed to use during the move
        :param critical_point: The critical point to move. In most situations
                               this is not needed. If not specified, the
                               current critical point will be moved. If
                               specified, the critical point must be one that
                               actually exists - that is, specifying
                               :py:attr:`.CriticalPoint.NOZZLE` when no pipette
                               is attached or :py:attr:`.CriticalPoint.TIP`
                               when no tip is applied will result in an error.
        :param max_speeds: An optional override for per-axis maximum speeds. If
                           an axis is specified, it will not move faster than
                           the given speed. Note that this does not make that
                           axis move precisely at the given speed; it only
                           it if it was going to go faster. Direct speed
                           is still set by ``speed``.
        """
        if not self._current_position:
            await self.home()

        mounts = self._mounts(mount)
        primary_mount = mounts[0]
        secondary_mount = None
        # Even with overloads, mypy cannot accept a length check on
        # a tuple to confirm whether there are one or two mounts
        # see: https://github.com/python/mypy/issues/1178
        if len(mounts) > 1:
            secondary_mount = mounts[1]  # type: ignore

        mount_offset = self._config.left_mount_offset
        if primary_mount == top_types.Mount.LEFT:
            primary_offset = top_types.Point(*mount_offset)
            s_offset = top_types.Point(0, 0, 0)
        else:
            primary_offset = top_types.Point(0, 0, 0)
            s_offset = top_types.Point(*mount_offset)

        if secondary_mount:
            primary_z = Axis.by_mount(primary_mount)
            secondary_z = Axis.by_mount(secondary_mount)
            primary_cp = self._critical_point_for(
                primary_mount, critical_point)
            s_cp = self._critical_point_for(
                secondary_mount, critical_point)
            target_position = OrderedDict(
                ((Axis.X, abs_position.x - primary_offset.x - primary_cp.x),
                 (Axis.Y, abs_position.y - primary_offset.y - primary_cp.y),
                 (primary_z, abs_position.z - primary_offset.z - primary_cp.z),
                 (secondary_z, abs_position.z - s_offset.z - s_cp.z)
                 ))
        else:
            primary_cp =\
                self._critical_point_for(primary_mount, critical_point)
            primary_z = Axis.by_mount(primary_mount)
            secondary_z = None
            target_position = OrderedDict(
                ((Axis.X, abs_position.x - primary_offset.x - primary_cp.x),
                 (Axis.Y, abs_position.y - primary_offset.y - primary_cp.y),
                 (primary_z, abs_position.z - primary_offset.z - primary_cp.z)
                 ))

        await self._cache_and_maybe_retract_mount(primary_mount)
        await self._move(
            target_position, speed=speed,
            max_speeds=max_speeds, secondary_z=secondary_z)

    async def move_rel(self, mount: Union[top_types.Mount, PipettePair],
                       delta: top_types.Point,
                       speed: float = None,
                       max_speeds: Dict[Axis, float] = None,
                       check_bounds: MotionChecks = MotionChecks.NONE):
        """ Move the critical point of the specified mount by a specified
        displacement in a specified direction, at the specified speed.
        'speed' sets the speed of all axes to the given value. So, if multiple
        axes are to be moved, they will do so at the same speed
        """
        if not self._current_position:
            await self.home()

        mounts = self._mounts(mount)
        primary_mount = mounts[0]
        secondary_mount = None
        # Even with overloads, mypy cannot accept a length check on
        # a tuple to confirm whether there are one or two mounts
        # see: https://github.com/python/mypy/issues/1178
        if len(mounts) > 1:
            secondary_mount = mounts[1]  # type: ignore

        if secondary_mount:
            primary_z = Axis.by_mount(primary_mount)
            secondary_z = Axis.by_mount(secondary_mount)
            target_position = OrderedDict(
                ((Axis.X, self._current_position[Axis.X] + delta.x),
                 (Axis.Y, self._current_position[Axis.Y] + delta.y),
                 (primary_z, self._current_position[primary_z] + delta.z),
                 (secondary_z, self._current_position[secondary_z] + delta.z))
            )
        else:
            z_axis = Axis.by_mount(primary_mount)
            secondary_z = None
            target_position = OrderedDict(
                ((Axis.X, self._current_position[Axis.X] + delta.x),
                 (Axis.Y, self._current_position[Axis.Y] + delta.y),
                 (z_axis, self._current_position[z_axis] + delta.z))
            )

        await self._cache_and_maybe_retract_mount(primary_mount)
        await self._move(
            target_position, speed=speed,
            max_speeds=max_speeds, secondary_z=secondary_z,
            check_bounds=check_bounds)

    async def _cache_and_maybe_retract_mount(self, mount: top_types.Mount):
        """ Retract the 'other' mount if necessary

        If `mount` does not match the value in :py:attr:`_last_moved_mount`
        (and :py:attr:`_last_moved_mount` exists) then retract the mount
        in :py:attr:`_last_moved_mount`. Also unconditionally update
        :py:attr:`_last_moved_mount` to contain `mount`.
        """
        if mount != self._last_moved_mount and self._last_moved_mount:
            await self.retract(self._last_moved_mount, 10)
        self._last_moved_mount = mount

    async def _move_plunger(self, mount: Union[top_types.Mount, PipettePair],
                            dist: Sequence[float], speed: float = None,
                            acquire_lock: bool = True):
        all_axes_pos = OrderedDict((
            (Axis.X, self._current_position[Axis.X]),
            (Axis.Y, self._current_position[Axis.Y]),
        ))
        plunger_pos = OrderedDict()
        mounts = self._mounts(mount)
        secondary_z = None
        for idx, m in enumerate(mounts):
            z = Axis.by_mount(m)
            plunger = Axis.of_plunger(m)
            all_axes_pos[z] = self._current_position[z]
            plunger_pos[plunger] = dist[idx]
            if idx == 1:
                secondary_z = z
        all_axes_pos.update(plunger_pos)
        await self._move(all_axes_pos, speed, False,
                         acquire_lock=acquire_lock,
                         secondary_z=secondary_z)

    async def _move_relative_n_axes(
            self, mount: Union[top_types.Mount, PipettePair],
            target: Sequence[float], speed: float = None):
        all_axes_pos = OrderedDict((
            (Axis.X, self._current_position[Axis.X] + target[0]),
            (Axis.Y, self._current_position[Axis.Y] + target[1])))
        mounts = self._mounts(mount)
        secondary_z = None
        for idx, m, in enumerate(mounts):
            z = Axis.by_mount(m)
            t_index = idx + 2
            all_axes_pos[z] = self._current_position[z] + target[t_index]
            if idx == 1:
                secondary_z = z
        await self._move(all_axes_pos, speed=speed, secondary_z=secondary_z)

    def _get_transformed(
            self,
            to_transform_primary: Tuple[float, ...],
            to_transform_secondary: Tuple[float, ...]
    ) -> Tuple[Tuple, Tuple]:
        # Type ignored below because linal.apply_transform (rightly) specifies
        # Tuple[float, float, float] and the implied type from
        # target_position.items() is (rightly) Tuple[float, ...] with unbounded
        # size; unfortunately, mypy can’t quite figure out the length check
        # above that makes this OK
        primary_transformed = linal.apply_transform(
            self.robot_calibration.deck_calibration.attitude,
            to_transform_primary)  # type: ignore
        secondary_transformed = linal.apply_transform(
            self.robot_calibration.deck_calibration.attitude,
            to_transform_secondary)  # type: ignore
        return primary_transformed, secondary_transformed

    async def _move(self, target_position: 'OrderedDict[Axis, float]',
                    speed: float = None, home_flagged_axes: bool = True,
                    max_speeds: Dict[Axis, float] = None,
                    acquire_lock: bool = True,
                    secondary_z: Axis = None,
                    check_bounds: MotionChecks = MotionChecks.NONE):
        """ Worker function to apply robot motion.

        Robot motion means the kind of motions that are relevant to the robot,
        i.e. only one pipette plunger and mount move at the same time, and an
        XYZ move in the coordinate frame of one of the pipettes.

        ``target_position`` should be an ordered dict (ordered by XYZABC)
        of deck calibrated values, containing any specified XY motion and
        at most one of a ZA or BC components. The frame in which to move
        is identified by the presence of (ZA) or (BC).
        """
        await self._wait_for_is_running()
        # Transform only the x, y, and (z or a) axes specified since this could
        # get the b or c axes as well
        to_transform_primary = tuple((tp
                                      for ax, tp in target_position.items()
                                      if (ax in Axis.gantry_axes()
                                          and ax != secondary_z)))
        if secondary_z:
            to_transform_secondary = tuple(
                (0, 0, target_position[secondary_z]))
        else:
            to_transform_secondary = tuple((0, 0, 0))
        # Pre-fill the dict we’ll send to the backend with the axes we don’t
        # need to transform
        smoothie_pos = {ax.name: pos for ax, pos in target_position.items()
                        if ax not in Axis.gantry_axes()}
        if len(to_transform_primary) != 3:
            self._log.error("Move derived {} axes to transform from {}"
                            .format(len(to_transform_primary),
                                    target_position))
            raise ValueError("Moves must specify either exactly an "
                             "x, y, and (z or a) or none of them")
        primary_transformed, secondary_transformed =\
            self._get_transformed(to_transform_primary, to_transform_secondary)
        transformed = (*primary_transformed, secondary_transformed[2])
        # Since target_position is an OrderedDict with the axes ordered by
        # (x, y, z, a, b, c), and we’ll only have one of a or z (as checked
        # by the len(to_transform) check above) we can use an enumerate to
        # fuse the specified axes and the transformed values back together.
        # While we do this iteration, we’ll also check axis bounds.
        bounds = self._backend.axis_bounds
        to_check = {ax: transformed[idx]
                    for idx, ax in enumerate(target_position.keys())
                    if ax in Axis.gantry_axes()}
        check_motion_bounds(
            to_check,
            target_position,
            bounds, check_bounds)
        smoothie_pos.update({ax.name: pos for ax, pos in to_check.items()})
        checked_maxes = max_speeds or {}
        str_maxes = {ax.name: val for ax, val in checked_maxes.items()}
        async with contextlib.AsyncExitStack() as stack:
            if acquire_lock:
                await stack.enter_async_context(self._motion_lock)
            try:
                self._backend.move(smoothie_pos, speed=speed,
                                   home_flagged_axes=home_flagged_axes,
                                   axis_max_speeds=str_maxes)
            except Exception:
                self._log.exception('Move failed')
                self._current_position.clear()
                raise
            else:
                self._current_position.update(target_position)

    def get_engaged_axes(self) -> Dict[Axis, bool]:
        """ Which axes are engaged and holding. """
        return {Axis[ax]: eng
                for ax, eng in self._backend.engaged_axes().items()}

    @property
    def engaged_axes(self):
        return self.get_engaged_axes()

    async def disengage_axes(self, which: List[Axis]):
        self._backend.disengage_axes([ax.name for ax in which])

    def _fast_home(
            self, axes: Sequence[str],
            margin: float) -> Dict[str, float]:
        converted_axes = ''.join(axes)
        return self._backend.fast_home(converted_axes, margin)

    async def retract(
            self,
            mount: Union[top_types.Mount, PipettePair],
            margin: float = 10):
        """ Pull the specified mount up to its home position.

        Works regardless of critical point or home status.
        """
        await self._wait_for_is_running()
        if isinstance(mount, PipettePair):
            primary_ax = Axis.by_mount(mount.primary).name.upper()
            secondary_ax = Axis.by_mount(mount.secondary).name.upper()
            smoothie_ax: Tuple[str, ...] = (primary_ax, secondary_ax)
        else:
            smoothie_ax = (Axis.by_mount(mount).name.upper(), )

        async with self._motion_lock:
            smoothie_pos = self._fast_home(smoothie_ax, margin)
            self._current_position = self._deck_from_smoothie(smoothie_pos)

    def _critical_point_for(
            self, mount: top_types.Mount,
            cp_override: CriticalPoint = None) -> top_types.Point:
        """ Return the current critical point of the specified mount.

        The mount's critical point is the position of the mount itself, if no
        pipette is attached, or the pipette's critical point (which depends on
        tip status).

        If `cp_override` is specified, and that critical point actually exists,
        it will be used instead. Invalid `cp_override`s are ignored.
        """
        pip = self._attached_instruments[mount]
        if pip is not None and cp_override != CriticalPoint.MOUNT:
            return pip.critical_point(cp_override)
        else:
            # This offset is required because the motor driver coordinate system is
            # configured such that the end of a p300 single gen1's tip is 0.
            return top_types.Point(0, 0, 30)

    # Gantry/frame (i.e. not pipette) config API
    def get_config(self) -> RobotConfig:
        """ Get the robot's configuration object.

        :returns .RobotConfig: The object.
        """
        return self._config

    def set_config(self, config: RobotConfig):
        """ Replace the currently-loaded config """
        self._config = config

    config = property(fget=get_config, fset=set_config)

    async def update_config(self, **kwargs):
        """ Update values of the robot's configuration.

        `kwargs` should contain keys of the robot's configuration. For
        instance, `update_config(log_level='debug)` would change the API
        server log level to :py:attr:`logging.DEBUG`.

        Documentation on keys can be found in the documentation for
        :py:class:`.RobotConfig`.
        """
        self._config = replace(self._config, **kwargs)  # type: ignore

    async def update_deck_calibration(self, new_transform):
        pass

    # Pipette action API
    async def prepare_for_aspirate(
            self, mount: Union[top_types.Mount, PipettePair],
            rate: float = 1.0):
        """
        Prepare the pipette for aspiration.

        This must happen after every :py:meth:`blow_out` and should probably be
        called before every :py:meth:`aspirate`, while the pipette tip is not
        immersed in a well. It ensures that the plunger is at the 0-volume
        position of the pipette if necessary (if not necessary, it does
        nothing).

        If :py:meth:`aspirate` is called immediately after :py:meth:`blow_out`,
        the plunger is left at the ``blow_out`` position, below the ``bottom``
        position, and moving the plunger up during :py:meth:`aspirate` is
        expected to aspirate liquid - :py:meth:`aspirate` is called once the
        pipette tip is already in the well. This will cause a subtle over
        aspiration. To make the problem more obvious, :py:meth:`aspirate` will
        raise an exception if this method has not previously been called.
        """
        instruments = self._instruments_for(mount)
        self._ready_for_tip_action(
            instruments, HardwareAction.PREPARE_ASPIRATE)

        with_zero = filter(lambda i: i[0].current_volume == 0, instruments)
        for instr in with_zero:
            speed = self._plunger_speed(
                instr[0],
                instr[0].blow_out_flow_rate, 'aspirate')
            bottom = (instr[0].config.bottom, )
            await self._move_plunger(instr[1], bottom, speed=(speed * rate))
            instr[0].ready_to_aspirate = True

    async def aspirate(self, mount: Union[top_types.Mount, PipettePair],
                       volume: float = None, rate: float = 1.0):
        """
        Aspirate a volume of liquid (in microliters/uL) using this pipette
        from the *current location*. If no volume is passed, `aspirate` will
        default to max available volume (after taking into account the volume
        already present in the tip).

        The function :py:meth:`prepare_for_aspirate` must be called prior to
        calling this function, while the tip is above the well. This ensures
        that the pipette tip is in the proper position at the bottom of the
        pipette to begin aspiration, and prevents subtle over-aspiration if
        an aspirate is done immediately after :py:meth:`blow_out`. If
        :py:meth:`prepare_for_aspirate` has not been called since the last
        call to :py:meth:`aspirate`, an exception will be raised.

        mount : Mount.LEFT or Mount.RIGHT
        volume : [float] The number of microliters to aspirate
        rate : [float] Set plunger speed for this aspirate, where
            speed = rate * aspirate_speed
        """
        instruments = self._instruments_for(mount)
        self._ready_for_tip_action(instruments, HardwareAction.ASPIRATE)
        plunger_currents = {
            Axis.of_plunger(instr[1]): instr[0].config.plunger_current
            for instr in instruments}

        if volume is None:
            mod_log.debug(
                "No aspirate volume defined. Aspirating up to "
                "max_volume for the pipette")
            asp_vol = tuple(instr[0].available_volume for instr in instruments)
        else:
            asp_vol = tuple(volume for instr in instruments)

        if all([vol == 0 for vol in asp_vol]):
            return
        elif 0 in asp_vol:
            raise PairedPipetteConfigValueError(
                "Cannot only aspirate from one pipette")

        for instr, vol in zip(instruments, asp_vol):
            assert instr[0].ok_to_add_volume(vol), \
                "Cannot aspirate more than pipette max volume"

        dist = tuple(self._plunger_position(
                     instr[0],
                     instr[0].current_volume + vol, 'aspirate')
                     for instr, vol in zip(instruments, asp_vol))
        speed = min(
            self._plunger_speed(
                instr[0], instr[0].aspirate_flow_rate * rate, 'aspirate')
            for instr in instruments)
        try:
            self._backend.set_active_current(plunger_currents)
            await self._move_plunger(mount, dist, speed=speed)
        except Exception:
            self._log.exception('Aspirate failed')
            for instr in instruments:
                instr[0].set_current_volume(0)
            raise
        else:
            for instr, vol in zip(instruments, asp_vol):
                instr[0].add_current_volume(vol)

    async def dispense(self, mount: Union[top_types.Mount, PipettePair],
                       volume: float = None, rate: float = 1.0):
        """
        Dispense a volume of liquid in microliters(uL) using this pipette
        at the current location. If no volume is specified, `dispense` will
        dispense all volume currently present in pipette

        mount : Mount.LEFT or Mount.RIGHT
        volume : [float] The number of microliters to dispense
        rate : [float] Set plunger speed for this dispense, where
            speed = rate * dispense_speed
        """
        instruments = self._instruments_for(mount)
        self._ready_for_tip_action(instruments, HardwareAction.DISPENSE)

        plunger_currents = {
            Axis.of_plunger(instr[1]): instr[0].config.plunger_current
            for instr in instruments}
        if volume is None:
            disp_vol = tuple(instr[0].current_volume for instr in instruments)
            mod_log.debug("No dispense volume specified. Dispensing all "
                          "remaining liquid ({}uL) from pipette".format
                          (disp_vol))
        else:
            disp_vol = tuple(volume for instr in instruments)

        # Ensure we don't dispense more than the current volume
        disp_vol = tuple(
            min(instr[0].current_volume, vol)
            for instr, vol in zip(instruments, disp_vol))

        if all([vol == 0 for vol in disp_vol]):
            return
        elif 0 in disp_vol:
            raise PairedPipetteConfigValueError(
                "Cannot only dispense from one pipette")

        dist = tuple(self._plunger_position(
                     instr[0],
                     instr[0].current_volume - vol, 'dispense')
                     for instr, vol in zip(instruments, disp_vol))
        speed = min(self._plunger_speed(instr[0],
                    instr[0].dispense_flow_rate * rate, 'dispense')
                    for instr in instruments)

        try:
            self._backend.set_active_current(plunger_currents)
            await self._move_plunger(mount, dist, speed=speed)
        except Exception:
            self._log.exception('Dispense failed')
            for instr in instruments:
                instr[0].set_current_volume(0)
            raise
        else:
            for instr, vol in zip(instruments, disp_vol):
                instr[0].remove_current_volume(vol)

    def _plunger_position(self, instr: Pipette, ul: float,
                          action: 'UlPerMmAction') -> float:
        mm = ul / instr.ul_per_mm(ul, action)
        position = mm + instr.config.bottom
        return round(position, 6)

    def _plunger_speed(
            self, instr: Pipette, ul_per_s: float,
            action: 'UlPerMmAction') -> float:
        mm_per_s = ul_per_s / instr.ul_per_mm(instr.config.max_volume, action)
        return round(mm_per_s, 6)

    def _plunger_flowrate(
            self, instr: Pipette, mm_per_s: float,
            action: 'UlPerMmAction') -> float:
        ul_per_s = mm_per_s * instr.ul_per_mm(instr.config.max_volume, action)
        return round(ul_per_s, 6)

    async def blow_out(self, mount: Union[top_types.Mount, PipettePair]):
        """
        Force any remaining liquid to dispense. The liquid will be dispensed at
        the current location of pipette
        """
        instruments = self._instruments_for(mount)
        self._ready_for_tip_action(instruments, HardwareAction.BLOWOUT)
        plunger_currents = {
            Axis.of_plunger(instr[1]): instr[0].config.plunger_current
            for instr in instruments}
        blow_out = tuple(instr[0].config.blow_out for instr in instruments)

        self._backend.set_active_current(plunger_currents)
        speed = max(self._plunger_speed(
            instr[0], instr[0].blow_out_flow_rate, 'dispense')
            for instr in instruments)
        try:
            await self._move_plunger(mount, blow_out, speed=speed)
        except Exception:
            self._log.exception('Blow out failed')
            raise
        finally:
            for instr in instruments:
                instr[0].set_current_volume(0)
                instr[0].ready_to_aspirate = False

    @overload
    def _instruments_for(
        self, mount: top_types.Mount) -> Tuple[PipetteHandlingData]: ...

    @overload
    def _instruments_for(
        self, mount: PipettePair) ->\
        Tuple[PipetteHandlingData, PipetteHandlingData]: ...

    def _instruments_for(self, mount):
        if isinstance(mount, PipettePair):
            primary_mount = mount.primary
            secondary_mount = mount.secondary
            instr1 = self._attached_instruments[primary_mount]
            instr2 = self._attached_instruments[secondary_mount]
            return ((instr1, primary_mount), (instr2, secondary_mount))
        else:
            primary_mount = mount
            instr1 = self._attached_instruments[primary_mount]
            return ((instr1, primary_mount), )

    def _ready_for_pick_up_tip(self, targets: Sequence[PipetteHandlingData]):
        for pipettes in targets:
            if not pipettes[0]:
                raise top_types.PipetteNotAttachedError(
                    f'No pipette attached to {pipettes[1].name} mount')
            if pipettes[0].has_tip:
                raise TipAttachedError(
                    'Cannot pick up tip with a tip attached')
            self._log.debug(f'Picking up tip on {pipettes[0].name}')

    def _ready_for_tip_action(
            self, targets: Sequence[PipetteHandlingData],
            action: HardwareAction):
        for pipettes in targets:
            if not pipettes[0]:
                raise top_types.PipetteNotAttachedError(
                    f'No pipette attached to {pipettes[1].name} mount')
            if not pipettes[0].has_tip:
                raise NoTipAttachedError(
                    f'Cannot perform {action} without a tip attached')
            if (action == HardwareAction.ASPIRATE and
                pipettes[0].current_volume == 0 and
                    not pipettes[0].ready_to_aspirate):
                raise RuntimeError('Pipette not ready to aspirate')
            self._log.debug(f'{action} on {pipettes[0].name}')

    async def pick_up_tip(self,
                          mount: Union[top_types.Mount, PipettePair],
                          tip_length: float,
                          presses: int = None,
                          increment: float = None):
        """
        Pick up tip from current location.

        This is achieved by attempting to move the instrument down by its
        `pick_up_distance`, in a series of presses. This distance is larger
        than the space available in the tip, so the stepper motor will
        eventually skip steps, which is resolved by homing afterwards. The
        pick up operation is done at a current specified in the pipette config,
        which is experimentally determined to skip steps at a level of force
        sufficient to provide a good seal between the pipette nozzle and tip
        while also avoiding attaching the tip so firmly that it can't be
        dropped later.

        If ``presses`` or ``increment`` is not specified (or is ``None``),
        their value is taken from the pipette configuration.
        """
        instruments = self._instruments_for(mount)
        self._ready_for_pick_up_tip(instruments)
        plunger_currents = {
            Axis.of_plunger(instr[1]): instr[0].config.plunger_current
            for instr in instruments}
        z_axis_currents = {
            Axis.by_mount(instr[1]): instr[0].config.pick_up_current
            for instr in instruments}

        self._backend.set_active_current(plunger_currents)
        # Initialize plunger to bottom position
        bottom_positions = tuple(
            instr[0].config.bottom for instr in instruments)
        await self._move_plunger(
            mount, bottom_positions)

        if not presses or presses < 0:
            all_presses = tuple(
                instr[0].config.pick_up_presses for instr in instruments)
            if len(all_presses) > 1 and all_presses[0] != all_presses[1]:
                raise PairedPipetteConfigValueError(
                    "Number of pipette pickups must match")
            checked_presses = all_presses[0]
        else:
            checked_presses = presses

        if not increment or increment < 0:
            check_incr = tuple(
                instr[0].config.pick_up_increment for instr in instruments)
        else:
            check_incr = tuple(increment for instr in instruments)

        pick_up_speed = min(
            instr[0].config.pick_up_speed for instr in instruments)
        # Press the nozzle into the tip <presses> number of times,
        # moving further by <increment> mm after each press
        for i in range(checked_presses):
            # move nozzle down into the tip
            with self._backend.save_current():
                self._backend.set_active_current(z_axis_currents)
                dist = tuple(-1.0 * instr[0].config.pick_up_distance
                             + -1.0 * incrt * i
                             for instr, incrt in zip(instruments, check_incr))
                target_pos = (0, 0, *dist)
                await self._move_relative_n_axes(
                    mount, target_pos, pick_up_speed)

            # move nozzle back up
            backup_pos = (0, 0, *tuple(-d for d in dist))
            await self._move_relative_n_axes(mount, backup_pos)
        for instr in instruments:
            instr[0].add_tip(tip_length=tip_length)
            instr[0].set_current_volume(0)

        # neighboring tips tend to get stuck in the space between
        # the volume chamber and the drop-tip sleeve on p1000.
        # This extra shake ensures those tips are removed
        if any(['pickupTipShake' in instr[0].config.quirks
                for instr in instruments]):
            await self._shake_off_tips_pick_up(mount)
            await self._shake_off_tips_pick_up(mount)

        # Here we add in the debounce distance for the switch as
        # a safety precaution
        retract_target = max(instr[0].config.pick_up_distance
                             + incrt * checked_presses + 2
                             for instr, incrt in zip(instruments, check_incr))

        await self.retract(mount, retract_target)

    def set_current_tiprack_diameter(
            self, mount: Union[top_types.Mount, PipettePair],
            tiprack_diameter: float):
        instruments = self._instruments_for(mount)
        for instr in instruments:
            assert instr[0]
            self._log.info(
                "Updating tip rack diameter on pipette mount: "
                f"{instr[1]}, tip diameter: {tiprack_diameter} mm")
            instr[0].current_tiprack_diameter = tiprack_diameter

    def set_working_volume(
            self, mount: Union[top_types.Mount, PipettePair],
            tip_volume: int):
        instruments = self._instruments_for(mount)
        for instr in instruments:
            assert instr[0]
            self._log.info(
                "Updating working volume on pipette mount:"
                f"{instr[1]}, tip volume: {tip_volume} ul")
            instr[0].working_volume = tip_volume

    async def drop_tip(
            self,
            mount: Union[top_types.Mount, PipettePair],
            home_after=True):
        """
        Drop tip at the current location

        :param Mount mount: The mount to drop a tip from
        :param bool home_after: Home the plunger motor after dropping tip. This
                                is used in case the plunger motor skipped while
                                dropping the tip, and is also used to recover
                                the ejector shroud after a drop.
        """

        instruments = self._instruments_for(mount)
        self._ready_for_tip_action(instruments, HardwareAction.DROPTIP)
        plunger_currents = {
            Axis.of_plunger(instr[1]): instr[0].config.plunger_current
            for instr in instruments}
        drop_tip_currents = {
            Axis.of_plunger(instr[1]): instr[0].config.drop_tip_current
            for instr in instruments}
        plunger_axes = tuple(
            Axis.of_plunger(instr[1]).name.upper() for instr in instruments)

        bottom = tuple(instr[0].config.bottom for instr in instruments)
        droptip = tuple(instr[0].config.drop_tip for instr in instruments)
        speed = min(instr[0].config.drop_tip_speed for instr in instruments)

        async def _drop_tip():
            self._backend.set_active_current(plunger_currents)
            await self._move_plunger(mount, bottom)
            self._backend.set_active_current(drop_tip_currents)
            await self._move_plunger(
                mount, droptip, speed=speed)
            if home_after:
                safety_margin = abs(max(bottom) - max(droptip))
                smoothie_pos = self._backend.fast_home(
                    plunger_axes, safety_margin)
                self._current_position = self._deck_from_smoothie(
                    smoothie_pos)
                self._backend.set_active_current(plunger_currents)
                await self._move_plunger(mount, bottom)

        if any(['doubleDropTip' in instr[0].config.quirks
                for instr in instruments]):
            await _drop_tip()
        await _drop_tip()

        if any(['dropTipShake' in instr[0].config.quirks
                for instr in instruments]):
            diameter = min(
                instr[0].current_tiprack_diameter
                for instr in instruments)
            await self._shake_off_tips_drop(mount, diameter)
        self._backend.set_active_current(plunger_currents)
        for instr in instruments:
            instr[0].set_current_volume(0)
            instr[0].current_tiprack_diameter = 0.0
            instr[0].remove_tip()

    async def _shake_off_tips_drop(self, mount, tiprack_diameter):
        # tips don't always fall off, especially if resting against
        # tiprack or other tips below it. To ensure the tip has fallen
        # first, shake the pipette to dislodge partially-sealed tips,
        # then second, raise the pipette so loosened tips have room to fall
        shake_off_dist = SHAKE_OFF_TIPS_DROP_DISTANCE
        if tiprack_diameter > 0.0:
            shake_off_dist = min(shake_off_dist, tiprack_diameter / 4)
        shake_off_dist = max(shake_off_dist, 1.0)

        shake_pos = top_types.Point(-shake_off_dist, 0, 0)  # move left
        await self.move_rel(mount, shake_pos, speed=SHAKE_OFF_TIPS_SPEED)
        shake_pos = top_types.Point(2 * shake_off_dist, 0, 0)    # move right
        await self.move_rel(mount, shake_pos, speed=SHAKE_OFF_TIPS_SPEED)
        shake_pos = top_types.Point(-shake_off_dist, 0, 0)  # original position
        await self.move_rel(mount, shake_pos, speed=SHAKE_OFF_TIPS_SPEED)
        # raise the pipette upwards so we are sure tip has fallen off
        up_pos = top_types.Point(0, 0, DROP_TIP_RELEASE_DISTANCE)
        await self.move_rel(mount, up_pos)

    async def _shake_off_tips_pick_up(
            self,
            mount: Union[top_types.Mount, PipettePair]):
        # tips don't always fall off, especially if resting against
        # tiprack or other tips below it. To ensure the tip has fallen
        # first, shake the pipette to dislodge partially-sealed tips,
        # then second, raise the pipette so loosened tips have room to fall
        shake_off_dist = SHAKE_OFF_TIPS_PICKUP_DISTANCE

        shake_pos = top_types.Point(-shake_off_dist, 0, 0)  # move left
        await self.move_rel(mount, shake_pos, speed=SHAKE_OFF_TIPS_SPEED)
        shake_pos = top_types.Point(2 * shake_off_dist, 0, 0)    # move right
        await self.move_rel(mount, shake_pos, speed=SHAKE_OFF_TIPS_SPEED)
        shake_pos = top_types.Point(-shake_off_dist, 0, 0)  # original position
        await self.move_rel(mount, shake_pos, speed=SHAKE_OFF_TIPS_SPEED)
        shake_pos = top_types.Point(0, -shake_off_dist, 0)  # move front
        await self.move_rel(mount, shake_pos, speed=SHAKE_OFF_TIPS_SPEED)
        shake_pos = top_types.Point(0, 2 * shake_off_dist, 0)    # move back
        await self.move_rel(mount, shake_pos, speed=SHAKE_OFF_TIPS_SPEED)
        shake_pos = top_types.Point(0, -shake_off_dist, 0)  # original position
        await self.move_rel(mount, shake_pos, speed=SHAKE_OFF_TIPS_SPEED)
        # raise the pipette upwards so we are sure tip has fallen off
        up_pos = top_types.Point(0, 0, DROP_TIP_RELEASE_DISTANCE)
        await self.move_rel(mount, up_pos)

    async def find_modules(
            self, by_model: modules.types.ModuleModel,
            resolved_type: modules.types.ModuleType
            ) -> Tuple[List[modules.AbstractModule], Optional[modules.AbstractModule]]:
        modules_result = await self._backend.module_controls.parse_modules(
            by_model, resolved_type)
        return modules_result

    # Pipette config api
    def calibrate_plunger(self,
                          mount: top_types.Mount,
                          top: float = None, bottom: float = None,
                          blow_out: float = None, drop_tip: float = None):
        """
        Set calibration values for the pipette plunger.
        This can be called multiple times as the user sets each value,
        or you can set them all at once.
        :param top: Touching but not engaging the plunger.
        :param bottom: Must be above the pipette's physical hard-stop, while
        still leaving enough room for 'blow_out'
        :param blow_out: Plunger is pushed down enough to expel all liquids.
        :param drop_tip: Position that causes the tip to be released from the
        pipette
        """
        instr = self._attached_instruments[mount]
        if not instr:
            raise top_types.PipetteNotAttachedError(
                "No pipette attached to {} mount".format(mount.name))

        pos_dict: Dict = {
            'top': instr.config.top,
            'bottom': instr.config.bottom,
            'blow_out': instr.config.blow_out,
            'drop_tip': instr.config.drop_tip}
        if top is not None:
            pos_dict['top'] = top
        if bottom is not None:
            pos_dict['bottom'] = bottom
        if blow_out is not None:
            pos_dict['blow_out'] = blow_out
        if bottom is not None:
            pos_dict['drop_tip'] = drop_tip
        for key in pos_dict.keys():
            instr.update_config_item(key, pos_dict[key])

    def set_flow_rate(self, mount,
                      aspirate=None, dispense=None, blow_out=None):
        this_pipette = self._attached_instruments[mount]
        if not this_pipette:
            raise top_types.PipetteNotAttachedError(
                "No pipette attached to {} mount".format(mount))
        if aspirate:
            this_pipette.aspirate_flow_rate = aspirate
        if dispense:
            this_pipette.dispense_flow_rate = dispense
        if blow_out:
            this_pipette.blow_out_flow_rate = blow_out

    def set_pipette_speed(self, mount,
                          aspirate=None, dispense=None, blow_out=None):
        this_pipette = self._attached_instruments[mount]
        if not this_pipette:
            raise top_types.PipetteNotAttachedError(
                "No pipette attached to {} mount".format(mount))
        if aspirate:
            this_pipette.aspirate_flow_rate = self._plunger_flowrate(
                this_pipette, aspirate, 'aspirate')
        if dispense:
            this_pipette.dispense_flow_rate = self._plunger_flowrate(
                this_pipette, dispense, 'dispense')
        if blow_out:
            this_pipette.blow_out_flow_rate = self._plunger_flowrate(
                this_pipette, blow_out, 'dispense')

    def get_instrument_max_height(
            self,
            mount: top_types.Mount,
            critical_point: CriticalPoint = None) -> float:
        """Return max achievable height of the attached instrument
        based on the current critical point
        """
        pip = self._attached_instruments[mount]
        assert pip
        cp = self._critical_point_for(mount, critical_point)

        max_height = pip.config.home_position - \
            self._config.z_retract_distance + cp.z

        return max_height

    def clean_up(self):
        """ Get the API ready to stop cleanly. """
        self._backend.clean_up()



"""Basic pipette data state and store."""
from dataclasses import dataclass
from typing import Dict, List, Mapping, Optional, Tuple
from typing_extensions import final

from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.types import MountType, Mount as HwMount

from .. import errors
from ..types import PipetteName
from ..commands import (
    CompletedCommandType,
    LoadPipetteResult,
    AspirateResult,
    DispenseResult,
)
from .substore import Substore, CommandReactive


@final
@dataclass(frozen=True)
class PipetteData:
    """Pipette state data."""

    mount: MountType
    pipette_name: PipetteName


@final
@dataclass(frozen=True)
class HardwarePipette:
    """Hardware pipette data."""

    mount: HwMount
    config: PipetteDict


class PipetteState:
    """Basic labware data state and getter methods."""

    _pipettes_by_id: Dict[str, PipetteData]
    _aspirated_volume_by_id: Dict[str, float]

    def __init__(self) -> None:
        """Initialize a PipetteState instance."""
        self._pipettes_by_id = {}
        self._aspirated_volume_by_id = {}

    def get_pipette_data_by_id(self, pipette_id: str) -> PipetteData:
        """Get pipette data by the pipette's unique identifier."""
        try:
            return self._pipettes_by_id[pipette_id]
        except KeyError:
            raise errors.PipetteDoesNotExistError(f"Pipette {pipette_id} not found.")

    def get_all_pipettes(self) -> List[Tuple[str, PipetteData]]:
        """Get a list of all pipette entries in state."""
        return [entry for entry in self._pipettes_by_id.items()]

    def get_pipette_data_by_mount(self, mount: MountType) -> Optional[PipetteData]:
        """Get pipette data by the pipette's mount."""
        for pipette in self._pipettes_by_id.values():
            if pipette.mount == mount:
                return pipette
        return None

    def get_hardware_pipette(
        self,
        pipette_id: str,
        attached_pipettes: Mapping[HwMount, Optional[PipetteDict]],
    ) -> HardwarePipette:
        """Get a pipette's hardware configuration and state by ID."""
        pipette_data = self.get_pipette_data_by_id(pipette_id)
        pipette_name = pipette_data.pipette_name
        mount = pipette_data.mount

        hw_mount = mount.to_hw_mount()
        hw_config = attached_pipettes[hw_mount]

        if hw_config is None:
            raise errors.PipetteNotAttachedError(f"No pipetted attached on {mount}")
        # TODO(mc, 2020-11-12): support hw_pipette.act_as
        elif hw_config["name"] != pipette_name:
            raise errors.PipetteNotAttachedError(
                f"Found {hw_config['name']} on {mount}, "
                f"but {pipette_id} is a {pipette_name}"
            )

        return HardwarePipette(mount=hw_mount, config=hw_config)

    def get_aspirated_volume(self, pipette_id: str) -> float:
        """Get the currently aspirated volume of a pipette by ID."""
        try:
            return self._aspirated_volume_by_id[pipette_id]
        except KeyError:
            raise errors.PipetteDoesNotExistError(
                f"Pipette {pipette_id} not found; unable to get current volume."
            )

    def get_is_ready_to_aspirate(
        self,
        pipette_id: str,
        pipette_config: PipetteDict,
    ) -> bool:
        """Get whether a pipette is ready to aspirate."""
        return (
            self.get_aspirated_volume(pipette_id) > 0
            or pipette_config["ready_to_aspirate"]
        )


class PipetteStore(Substore[PipetteState], CommandReactive):
    """Pipette state container."""

    _state: PipetteState

    def __init__(self) -> None:
        """Initialize a PipetteStore and its state."""
        self._state = PipetteState()

    def handle_completed_command(self, command: CompletedCommandType) -> None:
        """Modify state in reaction to a completed command."""
        if isinstance(command.result, LoadPipetteResult):
            pipette_id = command.result.pipetteId

            self._state._pipettes_by_id[pipette_id] = PipetteData(
                pipette_name=command.request.pipetteName, mount=command.request.mount
            )
            self._state._aspirated_volume_by_id[pipette_id] = 0

        elif isinstance(command.result, AspirateResult):
            pipette_id = command.request.pipetteId
            previous_volume = self._state._aspirated_volume_by_id[pipette_id]
            next_volume = previous_volume + command.result.volume

            self._state._aspirated_volume_by_id[pipette_id] = next_volume

        elif isinstance(command.result, DispenseResult):
            pipette_id = command.request.pipetteId
            previous_volume = self._state._aspirated_volume_by_id[pipette_id]
            next_volume = max(0, previous_volume - command.result.volume)

            self._state._aspirated_volume_by_id[pipette_id] = next_volume



from pathlib import Path

from opentrons.config import get_opentrons_path

OPENTRONS_NAMESPACE = 'opentrons'
CUSTOM_NAMESPACE = 'custom_beta'
STANDARD_DEFS_PATH = Path("labware/definitions/2")
USER_DEFS_PATH = get_opentrons_path('labware_user_definitions_dir_v2')

SHORT_TRASH_DECK = 'ot2_short_trash'
STANDARD_DECK = 'ot2_standard'



import logging
from typing import Dict

from opentrons.protocols.labware.definition import (
    IndexFileInformation, get_labware_definition)

from opentrons.calibration_storage import get
from opentrons.protocols.context.labware import \
    AbstractLabware
from opentrons.protocols.context.protocol_api.labware import \
    LabwareImplementation
from opentrons.types import Location
from opentrons_shared_data.labware.dev_types import LabwareDefinition


MODULE_LOG = logging.getLogger(__name__)


def load_from_definition(
        definition: LabwareDefinition,
        parent: Location,
        label: str = None) -> AbstractLabware:
    """
    Return a labware object constructed from a provided labware definition dict

    :param definition: A dict representing all required data for a labware,
        including metadata such as the display name of the labware, a
        definition of the order to iterate over wells, the shape of wells
        (shape, physical dimensions, etc), and so on. The correct shape of
        this definition is governed by the "labware-designer" project in
        the Opentrons/opentrons repo.
    :param parent: A :py:class:`.Location` representing the location where
                   the front and left most point of the outside of labware is
                   (often the front-left corner of a slot on the deck).
    :param str label: An optional label that will override the labware's
                      display name from its definition
    """
    labware = LabwareImplementation(
            definition=definition, parent=parent, label=label)
    index_info = IndexFileInformation.from_labware(labware)
    offset = get.get_labware_calibration(
        definition=index_info.definition,
        lookup_path=index_info.path,
        parent=index_info.parent)
    labware.set_calibration(offset)
    return labware


def load(
    load_name: str,
    parent: Location,
    label: str = None,
    namespace: str = None,
    version: int = 1,
    bundled_defs: Dict[str, LabwareDefinition] = None,
    extra_defs: Dict[str, LabwareDefinition] = None,
) -> AbstractLabware:
    """
    Return a labware object constructed from a labware definition dict looked
    up by name (definition must have been previously stored locally on the
    robot)

    :param load_name: A string to use for looking up a labware definition
        previously saved to disc. The definition file must have been saved in a
        known location
    :param parent: A :py:class:`.Location` representing the location where
                   the front and left most point of the outside of labware is
                   (often the front-left corner of a slot on the deck).
    :param str label: An optional label that will override the labware's
                      display name from its definition
    :param str namespace: The namespace the labware definition belongs to.
        If unspecified, will search 'opentrons' then 'custom_beta'
    :param int version: The version of the labware definition. If unspecified,
        will use version 1.
    :param bundled_defs: If specified, a mapping of labware names to labware
        definitions. Only the bundle will be searched for definitions.
    :param extra_defs: If specified, a mapping of labware names to labware
        definitions. If no bundle is passed, these definitions will also be
        searched.
    """
    definition = get_labware_definition(
        load_name, namespace, version,
        bundled_defs=bundled_defs,
        extra_defs=extra_defs)
    return load_from_definition(definition, parent, label)



from typing import Optional, List
from typing_extensions import Literal, TypedDict

State = Literal[
    'loaded', 'running', 'finished', 'stopped', 'paused', 'error', None]


class StateInfo(TypedDict, total=False):
    message: str
    #: A message associated with the state change by the system for display
    changedAt: float
    #: The time at which the state changed, relative to the startTime
    estimatedDuration: float
    #: If relevant for the state (e.g. 'paused' caused by a delay() call) the
    #: duration estimated for the state
    userMessage: str
    #: If provided by the mechanism that changed the state, a message from the
    #: user


class LastCommand(TypedDict):
    id: int
    handledAt: int


class Error(TypedDict):
    timestamp: int
    error: Exception


class SnapPayload(TypedDict):
    state: State
    stateInfo: StateInfo
    startTime: Optional[float]
    doorState: Optional[str]
    blocked: Optional[bool]
    errors: List[Error]
    lastCommand: Optional[LastCommand]


class Message(TypedDict):
    topic: Literal['session']
    payload: SnapPayload


class CommandShortId(TypedDict):
    level: int
    description: str
    id: int



from enum import Enum, auto
from typing import Type, TypeVar

Configuration = TypeVar('Configuration', bound='GenericConfiguration')


class GenericConfiguration(Enum):
    @classmethod
    def configuration_type(
            cls: Type[Configuration], config: str) -> 'Configuration':
        for m in cls.__members__.values():
            if m.name == config.upper():
                return m
        raise AttributeError(f'{config} not available')


class ThermocyclerConfiguration(GenericConfiguration):
    FULL = auto()
    SEMI = auto()

    def __str__(self):
        return self.name



""" protocol_api: The user-facing API for OT2 protocols.

This package defines classes and functions for access through a protocol to
control the OT2.

"""
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION, MIN_SUPPORTED_VERSION  # noqa: F401, E501
from . import labware
from .contexts import (ProtocolContext,
                       InstrumentContext,
                       PairedInstrumentContext,
                       TemperatureModuleContext,
                       MagneticModuleContext,
                       ThermocyclerContext)

__all__ = ['ProtocolContext',
           'InstrumentContext',
           'TemperatureModuleContext',
           'MagneticModuleContext',
           'ThermocyclerContext',
           'PairedInstrumentContext',
           'labware']



import abc
import asyncio
import logging
import re
from pkg_resources import parse_version
from typing import Mapping, Optional
from opentrons.config import IS_ROBOT, ROBOT_FIRMWARE_DIR
from opentrons.hardware_control.util import use_or_initialize_loop
from opentrons.drivers.rpi_drivers.types import USBPort
from ..execution_manager import ExecutionManager
from .types import BundledFirmware, UploadFunction, InterruptCallback, LiveData

mod_log = logging.getLogger(__name__)


class AbstractModule(abc.ABC):
    """ Defines the common methods of a module. """

    @classmethod
    @abc.abstractmethod
    async def build(cls,
                    port: str,
                    usb_port: USBPort,
                    execution_manager: ExecutionManager,
                    interrupt_callback: InterruptCallback = None,
                    simulating: bool = False,
                    loop: asyncio.AbstractEventLoop = None,
                    sim_model: str = None) \
            -> 'AbstractModule':
        """ Modules should always be created using this factory.

        This lets the (perhaps blocking) work of connecting to and initializing
        a module be in a place that can be async.
        """
        pass

    @abc.abstractmethod
    def __init__(self,
                 port: str,
                 usb_port: USBPort,
                 execution_manager: ExecutionManager,
                 simulating: bool = False,
                 loop: asyncio.AbstractEventLoop = None,
                 sim_model: str = None) -> None:
        self._port = port
        self._usb_port = usb_port
        self._loop = use_or_initialize_loop(loop)
        self._execution_manager = execution_manager
        self._device_info: Mapping[str, str]
        self._bundled_fw: Optional[BundledFirmware] = self.get_bundled_fw()

    def get_bundled_fw(self) -> Optional[BundledFirmware]:
        """ Get absolute path to bundled version of module fw if available. """
        if not IS_ROBOT:
            return None
        name_to_fw_file_prefix = {
            "tempdeck": "temperature-module", "magdeck": "magnetic-module"}
        name = self.name()
        file_prefix = name_to_fw_file_prefix.get(name, name)

        MODULE_FW_RE = re.compile(f'^{file_prefix}@v(.*)[.](hex|bin)$')
        for fw_resource in ROBOT_FIRMWARE_DIR.iterdir():  # type: ignore
            matches = MODULE_FW_RE.search(fw_resource.name)
            if matches:
                return BundledFirmware(version=matches.group(1),
                                       path=fw_resource)

        mod_log.info(f"no available fw file found for: {file_prefix}")
        return None

    def has_available_update(self) -> bool:
        """ Return whether a newer firmware file is available """
        if self._device_info and self._bundled_fw:
            device_version = parse_version(self._device_info['version'])
            available_version = parse_version(self._bundled_fw.version)
            return available_version > device_version
        return False

    async def wait_for_is_running(self):
        if not self.is_simulated:
            await self._execution_manager.wait_for_is_running()

    async def make_cancellable(self, task: asyncio.Task):
        self._execution_manager.register_cancellable_task(task)

    @abc.abstractmethod
    def deactivate(self):
        """ Deactivate the module. """
        pass

    @property
    @abc.abstractmethod
    def status(self) -> str:
        """ Return some string describing status. """
        pass

    @property
    @abc.abstractmethod
    def device_info(self) -> Mapping[str, str]:
        """ Return a dict of the module's static information (serial, etc)"""
        pass

    @property
    @abc.abstractmethod
    def live_data(self) -> LiveData:
        """ Return a dict of the module's dynamic information """
        pass

    @property
    @abc.abstractmethod
    def is_simulated(self) -> bool:
        """ True if >this is a simulated module. """
        pass

    @property
    @abc.abstractmethod
    def port(self) -> str:
        """ The virtual port where the module is connected. """
        pass

    @property
    @abc.abstractmethod
    def usb_port(self) -> USBPort:
        """ The physical port where the module is connected. """
        pass

    @abc.abstractmethod
    async def prep_for_update(self) -> str:
        """ Prepare for an update.

        By the time this coroutine completes, the hardware should be ready
        to take an update. This implicitly tears down the module instance;
        it does not need to be either working or recoverable after this
        coroutine completes.

        :returns str: The port we're running on.
        """
        pass

    @property
    @abc.abstractmethod
    def interrupt_callback(self) -> InterruptCallback:
        pass

    @property
    def bundled_fw(self):
        return self._bundled_fw

    @abc.abstractmethod
    def model(self) -> str:
        """ A name for this specific module, matching module defs """
        pass

    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        """ A shortname used for looking up firmware, among other things """
        pass

    @classmethod
    @abc.abstractmethod
    def bootloader(cls) -> UploadFunction:
        """ Method used to upload file to this module's bootloader. """
        pass

    def cleanup(self) -> None:
        """ Clean up the module instance.

        Clean up, i.e. stop pollers, disconnect serial, etc in preparation for
        object destruction.
        """
        pass



from .types import APIVersion

MAX_SUPPORTED_VERSION = APIVersion(2, 11)
#: The maximum supported protocol API version in this release

V2_MODULE_DEF_VERSION = APIVersion(2, 3)
#: The first API version in which we prefer the v2 module defs

MIN_SUPPORTED_VERSION = APIVersion(2, 0)
#: The minimum supported protocol API version in this release



import logging
from typing import TYPE_CHECKING
from opentrons.types import Point
from opentrons.protocols.execution.types import LoadedLabware, Instruments
from .execute_json_v3 import _get_well
if TYPE_CHECKING:
    from opentrons_shared_data.protocol.dev_types import (
        MoveToWellParams
    )

MODULE_LOG = logging.getLogger(__name__)


def _move_to_well(
        instruments: Instruments,
        loaded_labware: LoadedLabware,
        params: 'MoveToWellParams') -> None:
    pipette_id = params['pipette']
    pipette = instruments[pipette_id]

    well = _get_well(loaded_labware, params)

    offset = params.get('offset', {})
    offsetPoint = Point(
        offset.get('x', 0),
        offset.get('y', 0),
        offset.get('z', 0))

    pipette.move_to(
        well.bottom().move(offsetPoint),
        force_direct=params.get('forceDirect'),
        minimum_z_height=params.get('minimumZHeight'))



import asyncio
import logging
from typing import Union, Optional, List, Callable
from opentrons.drivers.rpi_drivers.types import USBPort
from ..execution_manager import ExecutionManager
from . import types, update, mod_abc
from opentrons.drivers.thermocycler.driver import (
    HOLD_TIME_FUZZY_SECONDS,
    SimulatingDriver,
    Thermocycler as ThermocyclerDriver)


MODULE_LOG = logging.getLogger(__name__)


class Thermocycler(mod_abc.AbstractModule):
    """
    Under development. API subject to change without a version bump
    """
    @classmethod
    async def build(cls,
                    port: str,
                    usb_port: USBPort,
                    execution_manager: ExecutionManager,
                    interrupt_callback: types.InterruptCallback = None,
                    simulating: bool = False,
                    loop: asyncio.AbstractEventLoop = None,
                    sim_model: str = None):
        """Build and connect to a Thermocycler
        """

        mod = cls(port=port,
                  usb_port=usb_port,
                  interrupt_callback=interrupt_callback,
                  simulating=simulating,
                  loop=loop,
                  execution_manager=execution_manager,
                  sim_model=sim_model)
        await mod._connect()
        return mod

    @classmethod
    def name(cls) -> str:
        return 'thermocycler'

    def model(self) -> str:
        return 'thermocyclerModuleV1'

    @classmethod
    def bootloader(cls) -> types.UploadFunction:
        return update.upload_via_bossa

    @staticmethod
    def _build_driver(
            simulating: bool,
            sim_model: str = None,
            interrupt_cb: Callable[[str], None] = None)\
            -> Union['SimulatingDriver', 'ThermocyclerDriver']:
        if simulating:
            return SimulatingDriver(sim_model=sim_model)
        else:
            return ThermocyclerDriver(interrupt_cb)

    def __init__(self,
                 port: str,
                 usb_port: USBPort,
                 execution_manager: ExecutionManager,
                 interrupt_callback: types.InterruptCallback = None,
                 simulating: bool = False,
                 loop: asyncio.AbstractEventLoop = None,
                 sim_model: str = None) -> None:
        super().__init__(port=port,
                         usb_port=usb_port,
                         simulating=simulating,
                         loop=loop,
                         execution_manager=execution_manager)
        self._driver: Union['SimulatingDriver', 'ThermocyclerDriver']
        self._interrupt_cb = interrupt_callback
        self._driver = self._build_driver(
            simulating,
            sim_model,
            interrupt_callback)

        self._total_cycle_count: Optional[int] = None
        self._current_cycle_index: Optional[int] = None
        self._total_step_count: Optional[int] = None
        self._current_step_index: Optional[int] = None

    def _clear_cycle_counters(self):
        self._total_cycle_count = None
        self._current_cycle_index = None
        self._total_step_count = None
        self._current_step_index = None

    async def deactivate_lid(self):
        """ Deactivate the lid heating pad"""
        await self.wait_for_is_running()
        return await self._driver.deactivate_lid()

    async def deactivate_block(self):
        """ Deactivate the block peltiers"""
        await self.wait_for_is_running()
        self._clear_cycle_counters()
        return await self._driver.deactivate_block()

    async def deactivate(self):
        """ Deactivate the block peltiers and lid heating pad"""
        await self.wait_for_is_running()
        self._clear_cycle_counters()
        return await self._driver.deactivate_all()

    async def open(self) -> str:
        """ Open the lid if it is closed"""
        await self.wait_for_is_running()
        return await self._driver.open()

    async def close(self) -> str:
        """ Close the lid if it is open"""
        await self.wait_for_is_running()
        return await self._driver.close()

    async def set_temperature(self, temperature,
                              hold_time_seconds: float = None,
                              hold_time_minutes: float = None,
                              ramp_rate: float = None,
                              volume: float = None):
        await self.wait_for_is_running()
        seconds = hold_time_seconds if hold_time_seconds is not None else 0
        minutes = hold_time_minutes if hold_time_minutes is not None else 0
        total_seconds = seconds + (minutes * 60)
        hold_time = total_seconds if total_seconds > 0 else 0
        await self._driver.set_temperature(temp=temperature,
                                           hold_time=hold_time,
                                           ramp_rate=ramp_rate,
                                           volume=volume)
        if hold_time:
            task = self._loop.create_task(
                self.wait_for_hold(hold_time))
        else:
            task = self._loop.create_task(
                self.wait_for_temp())
        await self.make_cancellable(task)
        await task

    async def _execute_cycle_step(self,
                                  step: types.ThermocyclerStep,
                                  index: int,
                                  volume: Optional[float]):
        await self.wait_for_is_running()
        self._current_step_index = index + 1  # science starts at 1
        temperature = step.get('temperature')
        hold_time_minutes = step.get('hold_time_minutes', None)
        hold_time_seconds = step.get('hold_time_seconds', None)
        ramp_rate = step.get('ramp_rate', None)
        await self.set_temperature(temperature=temperature,
                                   hold_time_minutes=hold_time_minutes,
                                   hold_time_seconds=hold_time_seconds,
                                   ramp_rate=ramp_rate,
                                   volume=volume)

    async def _execute_cycles(self,
                              steps: List[types.ThermocyclerStep],
                              repetitions: int,
                              volume: float = None):
        for rep in range(repetitions):
            self._current_cycle_index = rep + 1  # science starts at 1
            for step_idx, step in enumerate(steps):
                await self._execute_cycle_step(step, step_idx, volume)
                await self.wait_for_hold()

    async def cycle_temperatures(self,
                                 steps: List[types.ThermocyclerStep],
                                 repetitions: int,
                                 volume: float = None):
        await self.wait_for_is_running()
        self._total_cycle_count = repetitions
        self._total_step_count = len(steps)

        task = self._loop.create_task(
            self._execute_cycles(steps,
                                 repetitions,
                                 volume))
        await self.make_cancellable(task)
        await task

    async def set_lid_temperature(self, temperature: float):
        """ Set the lid temperature in deg Celsius """
        await self.wait_for_is_running()
        await self._driver.set_lid_temperature(temp=temperature)
        task = self._loop.create_task(self.wait_for_lid_temp())
        await self.make_cancellable(task)
        await task

    async def wait_for_lid_temp(self):
        """
        This method only exits if lid target temperature has been reached.

        Subject to change without a version bump.
        """
        if self.is_simulated:
            return

        while self._driver.lid_temp_status != 'holding at target':
            await asyncio.sleep(0.1)

    async def wait_for_temp(self):
        """
        This method only exits if set temperature has been reached.

        Subject to change without a version bump.
        """
        if self.is_simulated:
            return

        while self.status != 'holding at target':
            await asyncio.sleep(0.1)

    async def wait_for_hold(self, hold_time=0):
        """
        This method returns only when hold time has elapsed
        """
        if self.is_simulated:
            return

        # If hold time is within the HOLD_TIME_FUZZY_SECONDS time gap, then,
        # because of the driver's status poller delays, it is impossible to
        # know for certain if self.hold_time holds the most recent value.
        # So instead of counting on the cached self.hold_time, it is better to
        # just wait for hold_time time. (Skip if hold_time = 0 since we don't
        # want to wait in that case. Cached self.hold_time would be 0 anyway)
        if 0 < hold_time <= HOLD_TIME_FUZZY_SECONDS:
            await asyncio.sleep(hold_time)
        else:
            while self.hold_time != 0:
                await asyncio.sleep(0.1)

    @property
    def lid_target(self):
        return self._driver.lid_target

    @property
    def lid_temp(self):
        return self._driver.lid_temp

    @property
    def lid_status(self):
        return self._driver.lid_status

    @property
    def lid_temp_status(self):
        return self._driver.lid_temp_status

    @property
    def ramp_rate(self):
        return self._driver.ramp_rate

    @property
    def hold_time(self):
        return self._driver.hold_time

    @property
    def temperature(self):
        return self._driver.temperature

    @property
    def target(self):
        return self._driver.target

    @property
    def status(self):
        return self._driver.status

    @property
    def device_info(self):
        return self._device_info

    @property
    def total_cycle_count(self):
        return self._total_cycle_count

    @property
    def current_cycle_index(self):
        return self._current_cycle_index

    @property
    def total_step_count(self):
        return self._total_step_count

    @property
    def current_step_index(self):
        return self._current_step_index

    @property
    def live_data(self):
        return {
            'status': self.status,
            'data': {
                'lid': self.lid_status,
                'lidTarget': self.lid_target,
                'lidTemp': self.lid_temp,
                'currentTemp': self.temperature,
                'targetTemp': self.target,
                'holdTime': self.hold_time,
                'rampRate': self.ramp_rate,
                'currentCycleIndex': self.current_cycle_index,
                'totalCycleCount': self.total_cycle_count,
                'currentStepIndex': self.current_step_index,
                'totalStepCount': self.total_step_count,
            }
        }

    @property
    def is_simulated(self):
        return isinstance(self._driver, SimulatingDriver)

    @property
    def interrupt_callback(self):
        """ Fetch the current interrupt callback

        Exposes the interrupt callback used with the TCPoller, so it can be re-
        hooked in the new module instance after a firmware update.
        """
        return self._interrupt_cb

    @property
    def loop(self):
        return self._loop

    def set_loop(self, newLoop):
        self._loop = newLoop

    async def _connect(self):
        await self._driver.connect(self._port)
        self._device_info = await self._driver.get_device_info()

    @property
    def port(self):
        return self._port

    @property
    def usb_port(self) -> USBPort:
        return self._usb_port

    async def prep_for_update(self):
        await self._driver.enter_programming_mode()

        new_port = await update.find_bootloader_port()

        return new_port or self.port

    def cleanup(self) -> None:
        try:
            self._driver.disconnect()
        except Exception:
            MODULE_LOG.exception('Exception while cleaning up Thermocycler')

    def __del__(self):
        self.cleanup()



from typing import Any, Dict, NamedTuple, Optional, Union, TYPE_CHECKING
from dataclasses import dataclass
from .api_support.definitions import MIN_SUPPORTED_VERSION

if TYPE_CHECKING:
    from opentrons_shared_data.labware.dev_types import LabwareDefinition
    from opentrons_shared_data.protocol.dev_types import (
        JsonProtocol as JsonProtocolDef, Metadata as JsonProtocolMetadata
    )
    from .api_support.types import APIVersion

Metadata = Dict[str, Union[str, int]]


@dataclass(frozen=True)
class ProtocolCommon:
    text: str
    filename: Optional[str]
    api_level: 'APIVersion'
    metadata: Union[Metadata, 'JsonProtocolMetadata']


@dataclass(frozen=True)
class JsonProtocol(ProtocolCommon):
    schema_version: int
    contents: 'JsonProtocolDef'


@dataclass(frozen=True)
class PythonProtocol(ProtocolCommon):
    contents: Any  # This is the output of compile() which we can't type
    # these 'bundled_' attrs should only be included when the protocol is a zip
    bundled_labware: Optional[Dict[str, 'LabwareDefinition']]
    bundled_data: Optional[Dict[str, bytes]]
    bundled_python: Optional[Dict[str, str]]
    # this should only be included when the protocol is not a zip
    extra_labware: Optional[Dict[str, 'LabwareDefinition']]


Protocol = Union[JsonProtocol, PythonProtocol]


class BundleContents(NamedTuple):
    protocol: str
    bundled_labware: Dict[str, 'LabwareDefinition']
    bundled_data: Dict[str, bytes]
    bundled_python: Dict[str, str]


PROTOCOL_MALFORMED = """

A Python protocol for the OT2 must define a function called 'run' that takes a
single argument: the protocol context to call functions on. For instance, a run
function might look like this:

def run(ctx):
    ctx.comment('hello, world')

This function is called by the robot when the robot executes the protol.
This function is not present in the current protocol and must be added.
"""

PYTHON_API_VERSION_DEPRECATED = """

The python protocol you uploaded has the Python API Version {0}.  Robot server version 4.0.0 is
the official end of life of Python API Version {0}. The minimum supported Python API Version is {1}. This means that this protocol
will not run in robot server version 4.0.0 and above.
Please downgrade your robot server version if you wish to run this protocol. Otherwise, please upgrade this
protocol to Python API Version {1} or above.

To upgrade your protocol to Python API Version {1} or above, please view our documentation at https://docs.opentrons.com/v2/index.html.

Please contact support@opentrons.com to retrieve the previous software version and be guided
through the downgrade process.

"""  # noqa: E501


class MalformedProtocolError(Exception):
    def __init__(self, message):
        self.message = message
        super().__init__(message)

    def __str__(self):
        return self.message + PROTOCOL_MALFORMED

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.message)


class ApiDeprecationError(Exception):
    def __init__(self, version):
        self.version = version
        super().__init__(version)

    def __str__(self):
        return PYTHON_API_VERSION_DEPRECATED.format(
            self.version, MIN_SUPPORTED_VERSION)

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.version)



"""Synchronous ProtocolEngine client module."""
from uuid import uuid4
from typing import cast

from opentrons.types import MountType

from .. import commands
from ..state import StateView
from ..types import DeckSlotLocation, PipetteName, WellLocation
from .transports import AbstractSyncTransport


class SyncClient:
    """Synchronous Protocol Engine client."""

    def __init__(self, transport: AbstractSyncTransport) -> None:
        """Initialize the client with a transport."""
        self._transport = transport

    @property
    def state(self) -> StateView:
        """Get a view of the engine's state."""
        return self._transport.state

    @staticmethod
    def _create_command_id() -> str:
        return str(uuid4())

    def load_labware(
        self,
        location: DeckSlotLocation,
        load_name: str,
        namespace: str,
        version: int,
    ) -> commands.LoadLabwareResult:
        """Execute a LoadLabwareRequest and return the result."""
        request = commands.LoadLabwareRequest(
            location=location,
            loadName=load_name,
            namespace=namespace,
            version=version,
        )
        result = self._transport.execute_command(
            request=request,
            command_id=self._create_command_id(),
        )

        return cast(commands.LoadLabwareResult, result)

    def load_pipette(
        self,
        pipette_name: PipetteName,
        mount: MountType,
    ) -> commands.LoadPipetteResult:
        """Execute a LoadPipetteRequest and return the result."""
        request = commands.LoadPipetteRequest(
            pipetteName=pipette_name,
            mount=mount,
        )
        result = self._transport.execute_command(
            request=request,
            command_id=self._create_command_id(),
        )

        return cast(commands.LoadPipetteResult, result)

    def pick_up_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
    ) -> commands.PickUpTipResult:
        """Execute a PickUpTipRequest and return the result."""
        request = commands.PickUpTipRequest(
            pipetteId=pipette_id, labwareId=labware_id, wellName=well_name
        )
        result = self._transport.execute_command(
            request=request,
            command_id=self._create_command_id(),
        )

        return cast(commands.PickUpTipResult, result)

    def drop_tip(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
    ) -> commands.DropTipResult:
        """Execute a DropTipRequest and return the result."""
        request = commands.DropTipRequest(
            pipetteId=pipette_id, labwareId=labware_id, wellName=well_name
        )
        result = self._transport.execute_command(
            request=request, command_id=self._create_command_id()
        )
        return cast(commands.DropTipResult, result)

    def aspirate(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
        volume: float,
    ) -> commands.AspirateResult:
        """Execute an ``AspirateRequest``, returning the result."""
        request = commands.AspirateRequest(
            pipetteId=pipette_id,
            labwareId=labware_id,
            wellName=well_name,
            wellLocation=well_location,
            volume=volume,
        )
        result = self._transport.execute_command(
            request=request,
            command_id=self._create_command_id()
        )

        return cast(commands.AspirateResult, result)

    def dispense(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: WellLocation,
        volume: float,
    ) -> commands.DispenseResult:
        """Execute a ``DispenseRequest``, returning the result."""
        request = commands.DispenseRequest(
            pipetteId=pipette_id,
            labwareId=labware_id,
            wellName=well_name,
            wellLocation=well_location,
            volume=volume,
        )
        result = self._transport.execute_command(
            request=request, command_id=self._create_command_id()
        )
        return cast(commands.DispenseResult, result)



from __future__ import annotations
# this file defines types that require dev dependencies
# and are only relevant for static typechecking. this file should only
# be imported if typing.TYPE_CHECKING is True
import asyncio
from typing import Optional, Dict, Union

from typing_extensions import Protocol, TypedDict, Literal

from opentrons_shared_data.pipette.dev_types import (
    PipetteModel, PipetteName, ChannelCount
)

from opentrons.drivers.types import MoveSplit
from .types import HardwareEventType

from opentrons.types import Mount
from opentrons.config.pipette_config import PipetteConfig


class HasLoop(Protocol):
    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        ...


DoorStateNotificationType = Literal[HardwareEventType.DOOR_SWITCH_CHANGE]


class InstrumentSpec(TypedDict):
    model: Union[PipetteModel, None]
    id: Optional[str]


class AttachedInstrument(TypedDict):
    config: Optional[PipetteConfig]
    id: Optional[str]


AttachedInstruments = Dict[Mount, AttachedInstrument]


EIGHT_CHANNELS = Literal[8]
ONE_CHANNEL = Literal[1]


class PipetteDict(TypedDict):
    name: PipetteName
    model: PipetteModel
    pipette_id: str
    display_name: str
    min_volume: float
    max_volume: float
    channels: ChannelCount
    aspirate_flow_rate: float
    dispense_flow_rate: float
    blow_out_flow_rate: float
    aspirate_speed: float
    dispense_speed: float
    blow_out_speed: float
    current_volume: float
    tip_length: float
    working_volume: float
    tip_overlap: Dict[str, float]
    available_volume: float
    return_tip_height: float
    default_aspirate_flow_rates: Dict[str, float]
    default_dispense_flow_rates: Dict[str, float]
    default_blow_out_flow_rates: Dict[str,  float]
    default_aspirate_speeds: Dict[str, float]
    default_dispense_speeds: Dict[str, float]
    default_blow_out_speeds: Dict[str, float]
    ready_to_aspirate: bool
    has_tip: bool


class InstrumentHardwareConfigs(TypedDict):
    steps_per_mm: float
    home_pos: float
    max_travel: float
    idle_current: float
    splits: Optional[MoveSplit]



import asyncio
import logging
import threading
import time


log = logging.getLogger(__name__)


class ThreadedAsyncLock:
    """ A thread-safe async lock

    This is required to properly lock access to motion calls, which are
    a) done in async contexts (rpc methods and http methods) and should
       block as little as possible
    b) done from several different threads (rpc workers and main thread)

    This is a code wart that needs to be removed. It can be removed by
    - making smoothie async so we don't need worker threads anymore
    - removing said threads

    This object can be used as either an asynchronous context manager using
    ``async with`` or a synchronous context manager using ``with``.
    """

    def __init__(self):
        self._thread_lock = threading.Lock()

    def lock(self) -> '_Internal':
        """Create a context manager that locks access to a code block"""
        return _Internal(lock=self._thread_lock, forbid=False)

    def forbid(self) -> '_Internal':
        """Create a context manager that forbids concurrent attempts to
        access to a code block"""
        return _Internal(lock=self._thread_lock, forbid=True)


class ThreadedAsyncForbidden(Exception):
    """Exception indicating that lock is acquired and that blocking
    is forbidden"""

    def __init__(self, msg="Robot is currently moving. Please wait and try "
                           "this command again."):
        super().__init__(msg)


class _Internal:
    def __init__(self, lock: threading.Lock, forbid: bool):
        """
        The private context manager that interacts with the lock. It can
        behave in normal locking mode or in `forbid` mode. When `forbid` is
        True, then trying to acquire a lock that is already acquired will
        raise ThreadedAsyncForbidden. This forbids blocking on
         critical sections.

        :param lock: The Lock
        :param forbid: whether to block or raise an exception
        """
        self._thread_lock = lock
        self._forbid = forbid

    async def __aenter__(self):
        pref = f"[ThreadedAsyncLock tid {threading.get_ident()} "\
            f"task {asyncio.Task.current_task()}] "
        log.debug(pref + 'will acquire')
        then = time.perf_counter()
        while not self._thread_lock.acquire(blocking=False):
            if self._forbid:
                # Lock is already acquired and blocking is forbidden
                raise ThreadedAsyncForbidden()
            await asyncio.sleep(0.1)
        now = time.perf_counter()
        log.debug(pref + f'acquired in {now-then}s')

    async def __aexit__(self, exc_type, exc, tb):
        log.debug(f"[ThreadedAsyncLock tid {threading.get_ident()} "
                  f"task {asyncio.Task.current_task()}] will release")
        self._thread_lock.release()

    def __enter__(self):
        if not self._forbid:
            self._thread_lock.acquire()
        elif not self._thread_lock.acquire(blocking=False):
            # Lock is already acquired and blocking is forbidden
            raise ThreadedAsyncForbidden()

    def __exit__(self, exc_type, exc, tb):
        self._thread_lock.release()



"""Movement command handling."""
from typing import Optional
from opentrons.hardware_control.api import API as HardwareAPI

from ..types import DeckLocation, WellLocation
from ..state import StateView


class MovementHandler:
    """Implementation logic for gantry movement."""

    _state: StateView
    _hardware: HardwareAPI

    def __init__(
        self,
        state: StateView,
        hardware: HardwareAPI,
    ) -> None:
        """Initialize a MovementHandler instance."""
        self._state = state
        self._hardware = hardware

    async def move_to_well(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: Optional[WellLocation] = None,
        current_location: Optional[DeckLocation] = None,
    ) -> None:
        """Move to a specific well."""
        # get the pipette's mount and current critical point, if applicable
        pipette_location = self._state.motion.get_pipette_location(
            pipette_id=pipette_id,
            current_location=current_location,
        )

        hw_mount = pipette_location.mount.to_hw_mount()
        origin_cp = pipette_location.critical_point

        # get the origin of the movement from the hardware controller
        origin = await self._hardware.gantry_position(
            mount=hw_mount,
            critical_point=origin_cp,
        )
        max_travel_z = self._hardware.get_instrument_max_height(mount=hw_mount)

        # calculate the movement's waypoints
        waypoints = self._state.motion.get_movement_waypoints(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            origin=origin,
            origin_cp=origin_cp,
            max_travel_z=max_travel_z,
            current_location=current_location,
        )

        # move through the waypoints
        for wp in waypoints:
            await self._hardware.move_to(
                mount=hw_mount,
                abs_position=wp.position,
                critical_point=wp.critical_point
            )



"""Unique ID generation provider."""

from uuid import uuid4


class IdGenerator:
    """Unique ID generation provider."""

    def generate_id(self) -> str:
        """Generate a unique identifier.

        Uses UUIDv4 for safety in a multiprocessing environment.
        """
        return str(uuid4())



"""Load pipette command request, result, and implementation models."""
from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Field

from opentrons.types import MountType

from ..types import PipetteName
from .command import CommandImplementation, CommandHandlers


class LoadPipetteRequest(BaseModel):
    """A request to load a pipette on to a mount."""

    pipetteName: PipetteName = Field(
        ...,
        description="The name of the pipette to be required.",
    )
    mount: MountType = Field(
        ...,
        description="The mount the pipette should be present on.",
    )
    pipetteId: Optional[str] = Field(
        None,
        description="An optional ID to assign to this pipette. If None, an ID "
                    "will be generated."
    )

    def get_implementation(self) -> LoadPipetteImplementation:
        """Get the load pipette request's command implementation."""
        return LoadPipetteImplementation(self)


class LoadPipetteResult(BaseModel):
    """Result data for executing a LoadPipetteRequest."""

    pipetteId: str = Field(
        ...,
        description="An ID to reference this pipette in subsequent commands.",
    )


class LoadPipetteImplementation(
    CommandImplementation[LoadPipetteRequest, LoadPipetteResult]
):
    """Load pipette command implementation."""

    async def execute(self, handlers: CommandHandlers) -> LoadPipetteResult:
        """Check that requested pipette is attached and assign its identifier."""
        loaded_pipette = await handlers.equipment.load_pipette(
            pipette_name=self._request.pipetteName,
            mount=self._request.mount,
            pipette_id=self._request.pipetteId
        )

        return LoadPipetteResult(pipetteId=loaded_pipette.pipette_id)



from typing import Optional

from opentrons import types
from opentrons.protocols.api_support.types import APIVersion
from opentrons.hardware_control import CriticalPoint
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.protocols.api_support.definitions import MAX_SUPPORTED_VERSION
from opentrons.protocols.api_support.labware_like import LabwareLike
from opentrons.protocols.api_support.util import Clearances, build_edges, \
    FlowRates, PlungerSpeeds
from opentrons.protocols.geometry import planning
from opentrons.protocols.context.instrument import \
    AbstractInstrument
from opentrons.protocols.context.protocol import \
    AbstractProtocol
from opentrons.protocols.context.well import WellImplementation


class InstrumentContextImplementation(AbstractInstrument):
    """Implementation of the InstrumentContext interface."""

    _api_version: APIVersion
    _protocol_interface: AbstractProtocol
    _mount: types.Mount
    _instrument_name: str
    _default_speed: float
    _well_bottom_clearances: Clearances
    _flow_rates: FlowRates
    _speeds: PlungerSpeeds

    def __init__(self,
                 protocol_interface: AbstractProtocol,
                 mount: types.Mount,
                 instrument_name: str,
                 default_speed: float,
                 api_version: Optional[APIVersion] = None):
        """"Constructor"""
        # TODO AL 20201110 - Remove need for api_version in this module
        self._api_version = api_version or MAX_SUPPORTED_VERSION
        self._protocol_interface = protocol_interface
        self._mount = mount
        self._instrument_name = instrument_name
        self._default_speed = default_speed
        self._well_bottom_clearances = Clearances(
            default_aspirate=1.0,
            default_dispense=1.0
        )
        self._flow_rates = FlowRates(self)
        self._speeds = PlungerSpeeds(self)
        self._flow_rates.set_defaults(api_level=self._api_version)

    def get_default_speed(self) -> float:
        """Gets the speed at which the robot's gantry moves."""
        return self._default_speed

    def set_default_speed(self, speed: float) -> None:
        """Sets the speed at which the robot's gantry moves."""
        self._default_speed = speed

    def aspirate(self, volume: float, rate: float) -> None:
        """Aspirate a given volume of liquid from the specified location, using
        this pipette."""
        self._protocol_interface.get_hardware().hardware.aspirate(
            self._mount, volume, rate
        )

    def dispense(self, volume: float, rate: float) -> None:
        """Dispense a volume of liquid (in microliters/uL) using this pipette
        into the specified location."""
        self._protocol_interface.get_hardware().hardware.dispense(
            self._mount, volume, rate
        )

    def blow_out(self) -> None:
        """Blow liquid out of the tip."""
        self._protocol_interface.get_hardware().hardware.blow_out(self._mount)

    def touch_tip(self,
                  location: WellImplementation,
                  radius: float,
                  v_offset: float,
                  speed: float) -> None:
        """
        Touch the pipette tip to the sides of a well, with the intent of
        removing left-over droplets
        """
        # TODO al 20201110 - build_edges relies on where being a Well. This is
        #  an unpleasant compromise until refactoring build_edges to support
        #  WellImplementation.
        #  Also, build_edges should not require api_version.
        from opentrons.protocol_api.labware import Well

        edges = build_edges(
            where=Well(well_implementation=location),
            offset=v_offset,
            mount=self._mount,
            deck=self._protocol_interface.get_deck(),
            radius=radius,
            version=self._api_version
        )
        for edge in edges:
            self._protocol_interface.get_hardware().hardware.move_to(
                self._mount,
                edge,
                speed
            )

    def pick_up_tip(self,
                    well: WellImplementation,
                    tip_length: float,
                    presses: Optional[int],
                    increment: Optional[float]) -> None:
        """Pick up a tip for the pipette to run liquid-handling commands."""
        hw = self._protocol_interface.get_hardware().hardware
        geometry = well.get_geometry()

        hw.set_current_tiprack_diameter(
            self._mount, geometry.diameter)

        hw.pick_up_tip(
            self._mount,
            tip_length,
            presses,
            increment
        )
        hw.set_working_volume(
            self._mount, geometry.max_volume)

    def drop_tip(self, home_after: bool) -> None:
        """Drop the tip."""
        self._protocol_interface.get_hardware().hardware.drop_tip(
            self._mount,
            home_after=home_after
        )

    def home(self) -> None:
        """Home the mount"""
        self._protocol_interface.get_hardware().hardware.home_z(self._mount)
        self.home_plunger()

    def home_plunger(self) -> None:
        """Home the plunger associated with this mount."""
        self._protocol_interface.get_hardware().hardware.home_plunger(
            self._mount
        )

    def delay(self) -> None:
        """Delay protocol execution."""
        self._protocol_interface.delay(seconds=0, msg=None)

    def move_to(self,
                location: types.Location,
                force_direct: bool,
                minimum_z_height: Optional[float],
                speed: Optional[float]) -> None:
        """Move the instrument."""
        # prevent direct movement bugs in PAPI version >= 2.10
        location_cache_mount = (
            self._mount
            if self._api_version >= APIVersion(2, 10) else
            None
        )

        last_location = self._protocol_interface.get_last_location(
            mount=location_cache_mount
        )

        if last_location:
            from_lw = last_location.labware
        else:
            from_lw = LabwareLike(None)

        if not speed:
            speed = self.get_default_speed()

        hardware = self._protocol_interface.get_hardware().hardware

        from_center = from_lw.center_multichannel_on_wells() \
            if from_lw else False
        cp_override = CriticalPoint.XY_CENTER if from_center else None

        from_loc = types.Location(
            hardware.gantry_position(
                self._mount, critical_point=cp_override),
            from_lw)

        instr_max_height = hardware.get_instrument_max_height(self._mount)
        moves = planning.plan_moves(from_loc, location,
                                    self._protocol_interface.get_deck(),
                                    instr_max_height,
                                    force_direct=force_direct,
                                    minimum_z_height=minimum_z_height)

        try:
            for move in moves:
                hardware.move_to(
                    self._mount, move[0], critical_point=move[1], speed=speed,
                    max_speeds=self._protocol_interface.get_max_speeds().data)
        except Exception:
            self._protocol_interface.set_last_location(None)
            raise
        else:
            self._protocol_interface.set_last_location(
                location=location,
                mount=location_cache_mount
            )

    def get_mount(self) -> types.Mount:
        """Get the mount this pipette is attached to."""
        return self._mount

    def get_instrument_name(self) -> str:
        """Get the instrument name."""
        return self._instrument_name

    def get_pipette_name(self) -> str:
        """Get the pipette name."""
        return self.get_pipette()['name']

    def get_model(self) -> str:
        """Get the model name."""
        return self.get_pipette()['model']

    def get_min_volume(self) -> float:
        """Get the min volume."""
        return self.get_pipette()['min_volume']

    def get_max_volume(self) -> float:
        """Get the max volume."""
        return self.get_pipette()['max_volume']

    def get_current_volume(self) -> float:
        """Get the current volume."""
        return self.get_pipette()['current_volume']

    def get_available_volume(self) -> float:
        """Get the available volume."""
        return self.get_pipette()['available_volume']

    def get_pipette(self) -> PipetteDict:
        """Get the hardware pipette dictionary."""
        hw_manager = self._protocol_interface.get_hardware()
        pipette = hw_manager.hardware.get_attached_instrument(self._mount)
        if pipette is None:
            raise types.PipetteNotAttachedError()
        return pipette

    def get_channels(self) -> int:
        """Number of channels."""
        return self.get_pipette()['channels']

    def has_tip(self) -> bool:
        """Whether a tip is attached."""
        return self.get_pipette()['has_tip']

    def is_ready_to_aspirate(self) -> bool:
        return self.get_pipette()['ready_to_aspirate']

    def prepare_for_aspirate(self) -> None:
        self._protocol_interface.get_hardware(

        ).hardware.prepare_for_aspirate(self._mount)

    def get_return_height(self) -> float:
        """The height to return a tip to its tiprack."""
        return self.get_pipette().get('return_tip_height', 0.5)

    def get_well_bottom_clearance(self) -> Clearances:
        """The distance above the bottom of a well to aspirate or dispense."""
        return self._well_bottom_clearances

    def get_flow_rate(self) -> FlowRates:
        return self._flow_rates

    def get_speed(self) -> PlungerSpeeds:
        return self._speeds

    def set_flow_rate(
            self,
            aspirate: Optional[float] = None,
            dispense: Optional[float] = None,
            blow_out: Optional[float] = None) -> None:
        """Set the flow rates."""
        self._protocol_interface.get_hardware().hardware.set_flow_rate(
            mount=self._mount,
            aspirate=aspirate,
            dispense=dispense,
            blow_out=blow_out,
        )

    def set_pipette_speed(
            self,
            aspirate: Optional[float] = None,
            dispense: Optional[float] = None,
            blow_out: Optional[float] = None) -> None:
        """Set pipette speeds."""
        self._protocol_interface.get_hardware().hardware.set_pipette_speed(
            mount=self._mount,
            aspirate=aspirate,
            dispense=dispense,
            blow_out=blow_out,
        )



from typing import Callable, Dict, TYPE_CHECKING

from typing_extensions import Protocol, TypedDict

from opentrons_shared_data.protocol.dev_types import (
    BlowoutParams, DelayParams, PipetteAccessParams,
    StandardLiquidHandlingParams, TouchTipParams, MoveToSlotParams,
    TemperatureParams, ModuleIDParams, MagneticModuleEngageParams,
    ThermocyclerRunProfileParams, ThermocyclerSetTargetBlockParams,
    MoveToWellParams
)

if TYPE_CHECKING:
    from opentrons.protocol_api.protocol_context import ProtocolContext
    from opentrons.protocol_api.instrument_context import InstrumentContext
    from opentrons.protocol_api.module_contexts import (
        MagneticModuleContext, ThermocyclerContext, TemperatureModuleContext)
    from opentrons.protocol_api.labware import Labware


# this file defines types that require dev dependencies
# and are only relevant for static typechecking.
#
#  - code should be written so that this file can fail to import
#  - or the things defined in here can be None at execution time
#  - only types that match the above criteria should be put here
#  - please include this file as close to a leaf as possible

if Protocol is not None:
    class Dictable(Protocol):
        """ A protocol defining the _asdict interface to
        classes generated from typing.NamedTuple, which cannot
        be used as a type annotation because it is a type ctor
        not a type (https://github.com/python/mypy/issues/3915)
        """
        async def _asdict(self):
            ...


# using a lot of string literals here instead of the enum values from
# opentrons_shared_data.protocol.constants because of
# https://github.com/python/mypy/issues/4128
PipetteDispatch = TypedDict(
    'PipetteDispatch',
    {
        'delay': Callable[['ProtocolContext', 'DelayParams'], None],
        'blowout': Callable[
            [Dict[str, 'InstrumentContext'],
             Dict[str, 'Labware'],
             'BlowoutParams'], None],
        'pickUpTip': Callable[
            [Dict[str, 'InstrumentContext'],
             Dict[str, 'Labware'],
             'PipetteAccessParams'], None],
        'dropTip': Callable[
            [Dict[str, 'InstrumentContext'],
             Dict[str, 'Labware'],
             'PipetteAccessParams'], None],
        'aspirate': Callable[
            [Dict[str, 'InstrumentContext'],
             Dict[str, 'Labware'],
             'StandardLiquidHandlingParams'], None],
        'dispense': Callable[
            [Dict[str, 'InstrumentContext'],
             Dict[str, 'Labware'],
             'StandardLiquidHandlingParams'], None],
        'touchTip': Callable[
            [Dict[str, 'InstrumentContext'],
             Dict[str, 'Labware'],
             'TouchTipParams'], None],
        'moveToSlot': Callable[
            ['ProtocolContext',
             Dict[str, 'InstrumentContext'],
             'MoveToSlotParams'], None],
        'moveToWell': Callable[
            [Dict[str, 'InstrumentContext'],
             Dict[str, 'Labware'],
             'MoveToWellParams'], None],
        'airGap': Callable[
            [Dict[str, 'InstrumentContext'],
             Dict[str, 'Labware'],
             'StandardLiquidHandlingParams'], None],
    },
    total=False)


JsonV4MagneticModuleDispatch = TypedDict(
    'JsonV4MagneticModuleDispatch',
    {
        'magneticModule/engageMagnet': Callable[
            ['MagneticModuleContext', 'MagneticModuleEngageParams'], None],
        'magneticModule/disengageMagnet': Callable[
            ['MagneticModuleContext', 'ModuleIDParams'], None],
    })

JsonV4TemperatureModuleDispatch = TypedDict(
    'JsonV4TemperatureModuleDispatch',
    {
        'temperatureModule/setTargetTemperature': Callable[
            ['TemperatureModuleContext', 'TemperatureParams'], None],
        'temperatureModule/deactivate': Callable[
            ['TemperatureModuleContext', 'ModuleIDParams'], None],
        'temperatureModule/awaitTemperature': Callable[
            ['TemperatureModuleContext', 'TemperatureParams'], None]
    }
)

JsonV4ThermocyclerDispatch = TypedDict(
    'JsonV4ThermocyclerDispatch',
    {
        'thermocycler/closeLid': Callable[
            ['ThermocyclerContext', 'ModuleIDParams'], None],
        'thermocycler/openLid': Callable[
            ['ThermocyclerContext', 'ModuleIDParams'], None],
        'thermocycler/deactivateBlock': Callable[
            ['ThermocyclerContext', 'ModuleIDParams'], None],
        'thermocycler/deactivateLid': Callable[
            ['ThermocyclerContext', 'ModuleIDParams'], None],
        'thermocycler/setTargetBlockTemperature': Callable[
            ['ThermocyclerContext', 'ThermocyclerSetTargetBlockParams'],
            None],
        'thermocycler/setTargetLidTemperature': Callable[
            ['ThermocyclerContext', 'TemperatureParams'], None],
        'thermocycler/runProfile': Callable[
            ['ThermocyclerContext', 'ThermocyclerRunProfileParams'], None],
        'thermocycler/awaitBlockTemperature': Callable[
            ['ThermocyclerContext', 'TemperatureParams'], None],
        'thermocycler/awaitLidTemperature': Callable[
            ['ThermocyclerContext', 'TemperatureParams'], None],
        'thermocycler/awaitProfileComplete': Callable[
            ['ThermocyclerContext', 'ModuleIDParams'], None]
    }
)



"""Basic labware data state and store."""
from dataclasses import dataclass
from typing import Dict, List, Sequence, Tuple

from opentrons_shared_data.pipette.dev_types import LabwareUri
from typing_extensions import final

from opentrons.protocols.models import (
    LabwareDefinition,
    WellDefinition,
)
from opentrons.calibration_storage.helpers import uri_from_details

from .. import errors
from ..resources import DeckFixedLabware
from ..commands import (
    CompletedCommandType,
    LoadLabwareResult,
    AddLabwareDefinitionResult,
)
from ..types import LabwareLocation, Dimensions
from .substore import Substore, CommandReactive


@final
@dataclass(frozen=True)
class LabwareData:
    """Labware data entry."""

    location: LabwareLocation
    uri: LabwareUri
    calibration: Tuple[float, float, float]


class LabwareState:
    """Basic labware data state and getter methods."""

    _labware_by_id: Dict[str, LabwareData]
    _labware_definitions_by_uri: Dict[str, LabwareDefinition]

    def __init__(self, deck_fixed_labware: Sequence[DeckFixedLabware]) -> None:
        """Initialize a LabwareState instance."""
        self._labware_definitions_by_uri = {
            uri_from_details(
                load_name=fixed_labware.definition.parameters.loadName,
                namespace=fixed_labware.definition.namespace,
                version=fixed_labware.definition.version
            ): fixed_labware.definition for fixed_labware in deck_fixed_labware
        }
        self._labware_by_id = {
            fixed_labware.labware_id: LabwareData(
                location=fixed_labware.location,
                uri=uri_from_details(
                    load_name=fixed_labware.definition.parameters.loadName,
                    namespace=fixed_labware.definition.namespace,
                    version=fixed_labware.definition.version
                ),
                calibration=(0, 0, 0),
            )
            for fixed_labware in deck_fixed_labware
        }

    def get_labware_data_by_id(self, labware_id: str) -> LabwareData:
        """Get labware data by the labware's unique identifier."""
        try:
            return self._labware_by_id[labware_id]
        except KeyError:
            raise errors.LabwareDoesNotExistError(f"Labware {labware_id} not found.")

    def get_labware_definition(self, labware_id: str) -> LabwareDefinition:
        """Get labware definition by the labware's unique identifier."""
        return self.get_definition_by_uri(
            self.get_labware_data_by_id(labware_id).uri
        )

    def get_definition_by_uri(self, uri: LabwareUri) -> LabwareDefinition:
        """Get the labware definition matching loadName namespace and version."""
        try:
            return self._labware_definitions_by_uri[uri]
        except KeyError:
            raise errors.LabwareDefinitionDoesNotExistError(
                f"Labware definition for matching {uri} not found.")

    def get_labware_location(self, labware_id: str) -> LabwareLocation:
        """Get labware location by the labware's unique identifier."""
        return self.get_labware_data_by_id(labware_id).location

    def get_all_labware(self) -> List[Tuple[str, LabwareData]]:
        """Get a list of all labware entries in state."""
        return [entry for entry in self._labware_by_id.items()]

    def get_labware_has_quirk(self, labware_id: str, quirk: str) -> bool:
        """Get if a labware has a certain quirk."""
        return quirk in self.get_quirks(labware_id=labware_id)

    def get_quirks(self, labware_id: str) -> List[str]:
        """Get a labware's quirks."""
        definition = self.get_labware_definition(labware_id)
        return definition.parameters.quirks or []

    def get_well_definition(
        self,
        labware_id: str,
        well_name: str,
    ) -> WellDefinition:
        """Get a well's definition by labware and well identifier."""
        definition = self.get_labware_definition(labware_id)

        try:
            return definition.wells[well_name]
        except KeyError:
            raise errors.WellDoesNotExistError(
                f"{well_name} does not exist in {labware_id}."
            )

    def get_tip_length(self, labware_id: str) -> float:
        """Get the tip length of a tip rack."""
        definition = self.get_labware_definition(labware_id)
        if definition.parameters.tipLength is None:
            raise errors.LabwareIsNotTipRackError(
                f"Labware {labware_id} has no tip length defined."
            )
        return definition.parameters.tipLength

    def get_definition_uri(self, labware_id: str) -> str:
        """Get a labware's definition URI."""
        return self.get_labware_data_by_id(labware_id).uri

    def is_tiprack(self, labware_id: str) -> bool:
        """Get whether labware is a tiprack."""
        definition = self.get_labware_definition(labware_id)
        return definition.parameters.isTiprack

    def get_load_name(self, labware_id: str) -> str:
        """Get the labware's load name."""
        definition = self.get_labware_definition(labware_id)
        return definition.parameters.loadName

    def get_dimensions(self, labware_id: str) -> Dimensions:
        """Get the labware's dimensions."""
        definition = self.get_labware_definition(labware_id)
        dims = definition.dimensions

        return Dimensions(
            x=dims.xDimension,
            y=dims.yDimension,
            z=dims.zDimension,
        )


class LabwareStore(Substore[LabwareState], CommandReactive):
    """Labware state container."""

    _state: LabwareState

    def __init__(self, deck_fixed_labware: Sequence[DeckFixedLabware]) -> None:
        """Initialize a labware store and its state."""
        self._state = LabwareState(deck_fixed_labware=deck_fixed_labware)

    def handle_completed_command(self, command: CompletedCommandType) -> None:
        """Modify state in reaction to a completed command."""
        if isinstance(command.result, LoadLabwareResult):
            uri = uri_from_details(
                namespace=command.result.definition.namespace,
                load_name=command.result.definition.parameters.loadName,
                version=command.result.definition.version
            )
            self._state._labware_definitions_by_uri[uri] = \
                command.result.definition
            self._state._labware_by_id[command.result.labwareId] = LabwareData(
                location=command.request.location,
                uri=uri,
                calibration=command.result.calibration,
            )
        elif isinstance(command.result, AddLabwareDefinitionResult):
            uri = uri_from_details(
                namespace=command.result.namespace,
                load_name=command.result.loadName,
                version=command.result.version
            )
            self._state._labware_definitions_by_uri[uri] =\
                command.request.definition



from __future__ import annotations
import asyncio
import functools
import logging
from copy import copy
from typing import Any, cast, Callable, Optional, Union, Set, TypeVar
from typing_extensions import TypedDict, Literal, Final

from opentrons.config import feature_flags as ff
from opentrons.broker import Broker
from opentrons.types import Point, Mount, Location
from opentrons.protocol_api import labware
from opentrons.hardware_control import CriticalPoint, ThreadedAsyncLock
from opentrons.hardware_control.types import OutOfBoundsMove, MotionChecks
from opentrons.hardware_control.adapters import SynchronousAdapter

from .models import Container, Instrument
from .util import robot_is_busy, RobotBusy


log = logging.getLogger(__name__)

State = Union[Literal['moving'], Literal['ready']]
VALID_STATES: Set[State] = {'moving', 'ready'}


def _well0(cont: labware.Labware) -> labware.Well:
    return cont.wells()[0]


Func = TypeVar('Func', bound=Callable)


def _home_if_first_call(func: Func) -> Func:
    """ Decorator to make a function home if it is the first one called in
    this session."""
    @functools.wraps(func)
    def decorated(*args: Any, **kwargs: Any) -> Any:
        self = args[0]
        if not self._has_homed:
            log.info("this is the first calibration action, homing")
            self._hardware.home()
            self._has_homed = True
        return func(*args, **kwargs)
    return cast(Func, decorated)


class Message(TypedDict):
    topic: Literal['calibration']
    name: Literal['state']
    payload: CalibrationManager


class CalibrationManager(RobotBusy):
    """
    Serves endpoints that are primarily used in
    opentrons/app/ui/robot/api-client/client.js
    """
    TOPIC: Final = 'calibration'

    def __init__(
            self,
            hardware: SynchronousAdapter,
            loop: asyncio.AbstractEventLoop = None,
            broker: Broker = None,
            lock: ThreadedAsyncLock = None) -> None:
        self._broker = broker or Broker()
        self._hardware = hardware
        self._loop = loop
        self.state = 'ready'
        self._lock = lock or ThreadedAsyncLock()
        self._has_homed = False

    @property
    def busy_lock(self) -> ThreadedAsyncLock:
        return self._lock

    def _set_state(self, state: State) -> None:
        if state not in VALID_STATES:
            raise ValueError(
                'State {0} not in {1}'.format(state, VALID_STATES))
        self.state = state
        self._on_state_changed()

    @robot_is_busy
    @_home_if_first_call
    def tip_probe(self, instrument: Instrument) -> None:
        log.warning("Deprecated call to tip_probe, update app")
        self._set_state('ready')

    @robot_is_busy
    @_home_if_first_call
    def pick_up_tip(self, instrument: Instrument, container: Container) -> None:
        if not isinstance(container, Container):
            raise ValueError(
                'Invalid object type {0}. Expected models.Container'
                .format(type(container)))

        inst = instrument._instrument
        log.info('Picking up tip from {} in {} with {}'.format(
            container.name, container.slot, instrument.name))
        self._set_state('moving')
        with instrument._context.temp_connect(self._hardware):
            loc = _well0(container._container)
            instrument._context.location_cache =\
                Location(self._hardware.gantry_position(
                    Mount[inst.mount.upper()],
                    critical_point=CriticalPoint.NOZZLE,
                    refresh=True),
                    loc)
            loc_leg = _well0(container._container)
            inst.pick_up_tip(loc_leg)
        self._set_state('ready')

    @robot_is_busy
    @_home_if_first_call
    def drop_tip(self, instrument: Instrument, container: Container) -> None:
        if not isinstance(container, Container):
            raise ValueError(
                'Invalid object type {0}. Expected models.Container'
                .format(type(container)))

        inst = instrument._instrument
        log.info('Dropping tip from {} in {} with {}'.format(
            container.name, container.slot, instrument.name))
        self._set_state('moving')
        with instrument._context.temp_connect(self._hardware):
            instrument._context.location_cache = None
            inst.drop_tip(_well0(container._container))
        self._set_state('ready')

    @robot_is_busy
    @_home_if_first_call
    def return_tip(self, instrument: Instrument) -> None:
        inst = instrument._instrument
        log.info('Returning tip from {}'.format(instrument.name))
        self._set_state('moving')
        with instrument._context.temp_connect(self._hardware):
            instrument._context.location_cache = None
            inst.return_tip()
        self._set_state('ready')

    @robot_is_busy
    @_home_if_first_call
    def move_to_front(self, instrument: Instrument) -> None:
        """Public face of move_to_front"""
        self._move_to_front(instrument)

    def _move_to_front(self, instrument: Instrument) -> None:
        """Private move_to_front that can be called internally"""
        inst = instrument._instrument
        log.info('Moving {}'.format(instrument.name))
        self._set_state('moving')
        current = self._hardware.gantry_position(
            Mount[inst.mount.upper()],
            critical_point=CriticalPoint.NOZZLE,
            refresh=True)
        dest = instrument._context.deck.position_for(5) \
            .point._replace(z=150)
        self._hardware.move_to(Mount[inst.mount.upper()],
                               current,
                               critical_point=CriticalPoint.NOZZLE)
        self._hardware.move_to(Mount[inst.mount.upper()],
                               dest._replace(z=current.z),
                               critical_point=CriticalPoint.NOZZLE)
        self._hardware.move_to(Mount[inst.mount.upper()],
                               dest, critical_point=CriticalPoint.NOZZLE)
        self._set_state('ready')

    @robot_is_busy
    @_home_if_first_call
    def move_to(self, instrument: Instrument, container: Container) -> None:
        if not isinstance(container, Container):
            raise ValueError(
                'Invalid object type {0}. Expected models.Container'
                .format(type(container)))

        inst = instrument._instrument
        cont = container._container
        target = _well0(cont).top()

        log.info('Moving {} to {} in {}'.format(
            instrument.name, container.name, container.slot))
        self._set_state('moving')

        with instrument._context.temp_connect(self._hardware):
            instrument._context.location_cache = None
            inst.move_to(target)

        self._set_state('ready')

    @robot_is_busy
    @_home_if_first_call
    def jog(self, instrument: Instrument, distance: float, axis: str) -> None:
        inst = instrument._instrument
        log.info('Jogging {} by {} in {}'.format(
            instrument.name, distance, axis))
        self._set_state('moving')
        try:
            self._hardware.move_rel(
                Mount[inst.mount.upper()], Point(**{axis: distance}),
                check_bounds=MotionChecks.HIGH)
        except OutOfBoundsMove:
            log.exception('Out of bounds jog')
        self._set_state('ready')

    @robot_is_busy
    @_home_if_first_call
    def home(self, instrument: Instrument) -> None:
        inst = instrument._instrument
        log.info('Homing {}'.format(instrument.name))
        self._set_state('moving')
        with instrument._context.temp_connect(self._hardware):
            instrument._context.location_cache = None
            inst.home()
        self._set_state('ready')

    @robot_is_busy
    def home_all(self, instrument: Instrument) -> None:
        # NOTE: this only takes instrument as a param, because we need
        # its reference to the ProtocolContext. This is code smell that
        # will be removed once sessions are managed better
        log.info('Homing via Calibration Manager')
        self._set_state('moving')
        with instrument._context.temp_connect(self._hardware):
            instrument._context.home()
        self._set_state('ready')

    @robot_is_busy
    def update_container_offset(
            self, container: Container, instrument: Instrument) -> None:
        inst = instrument._instrument
        log.info('Updating {} in {}'.format(container.name, container.slot))
        if 'centerMultichannelOnWells' in container._container.quirks:
            cp: Optional[CriticalPoint] = CriticalPoint.XY_CENTER
        else:
            cp = None
        here = self._hardware.gantry_position(Mount[inst.mount.upper()],
                                              critical_point=cp,
                                              refresh=True)
        # Reset calibration so we don’t actually calibrate the offset
        # relative to the old calibration
        container._container.set_calibration(Point(0, 0, 0))
        if ff.calibrate_to_bottom() and not (
                container._container.is_tiprack):
            orig = _well0(container._container).geometry.bottom()
        else:
            orig = _well0(container._container).geometry.top()
        delta = here - orig
        labware.save_calibration(container._container, delta)

    def _snapshot(self) -> Message:
        return {
            'topic': CalibrationManager.TOPIC,
            'name': 'state',
            'payload': copy(self)
        }

    def _on_state_changed(self) -> None:
        self._broker.publish(CalibrationManager.TOPIC, self._snapshot())



import asyncio
from typing import Set
from .types import ExecutionState, ExecutionCancelledError


class ExecutionManager():
    """ This class holds onto a flag that is up/set while the hardware
    hardware controller is running and down/cleared when the hardware
    controller is "paused".

    It also handles loop clean up through its cancel method.
    """
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self._state: ExecutionState = ExecutionState.RUNNING
        self._condition = asyncio.Condition(loop=loop)
        self._loop = loop
        self._cancellable_tasks: Set[asyncio.Task] = set()

    async def pause(self):
        async with self._condition:
            if self._state is ExecutionState.CANCELLED:
                raise ExecutionCancelledError
            else:
                self._state = ExecutionState.PAUSED

    async def resume(self):
        async with self._condition:
            if self._state is ExecutionState.CANCELLED:
                pass
            else:
                self._state = ExecutionState.RUNNING
                self._condition.notify_all()

    async def cancel(self):
        async with self._condition:
            self._state = ExecutionState.CANCELLED
            self._condition.notify_all()
            running_task = asyncio.current_task(self._loop)
            for t in self._cancellable_tasks:
                if t is not running_task:
                    t.cancel()

    async def reset(self):
        async with self._condition:
            self._state = ExecutionState.RUNNING
            self._condition.notify_all()

    async def get_state(self) -> ExecutionState:
        async with self._condition:
            return self._state

    def register_cancellable_task(self, task: asyncio.Task):
        self._cancellable_tasks.add(task)
        task.add_done_callback(lambda t: self._cancellable_tasks.discard(t))
        return task

    async def wait_for_is_running(self):
        async with self._condition:
            if self._state is ExecutionState.PAUSED:
                await self._condition.wait()
                if self._state is ExecutionState.CANCELLED:
                    raise ExecutionCancelledError
            elif self._state is ExecutionState.CANCELLED:
                raise ExecutionCancelledError
            else:
                pass



"""Motion state store and getters."""
from dataclasses import dataclass
from typing import List, Optional

from opentrons.types import MountType, Point
from opentrons.hardware_control.types import CriticalPoint
from opentrons.motion_planning import (
    MoveType,
    Waypoint,
    MotionPlanningError,
    get_waypoints,
)

from .. import commands, errors
from ..types import WellLocation, DeckLocation
from .substore import Substore, CommandReactive
from .labware import LabwareStore
from .pipettes import PipetteStore
from .geometry import GeometryStore


@dataclass(frozen=True)
class PipetteLocationData:
    """Pipette data used to determine the current gantry position."""

    mount: MountType
    critical_point: Optional[CriticalPoint]


class MotionState:
    """Motion planning state and getter methods."""

    _labware_store: LabwareStore
    _pipette_store: PipetteStore
    _geometry_store: GeometryStore
    _current_location: Optional[DeckLocation]

    def __init__(
        self,
        labware_store: LabwareStore,
        pipette_store: PipetteStore,
        geometry_store: GeometryStore,
    ) -> None:
        """Initialize a MotionState instance."""
        self._labware_store = labware_store
        self._pipette_store = pipette_store
        self._geometry_store = geometry_store
        self._current_location = None

    def get_current_deck_location(self) -> Optional[DeckLocation]:
        """Get the current pipette and deck location the protocol is at."""
        return self._current_location

    def get_pipette_location(
        self,
        pipette_id: str,
        current_location: Optional[DeckLocation] = None,
    ) -> PipetteLocationData:
        """Get the critical point of a pipette given the current location."""
        current_loc = current_location or self.get_current_deck_location()
        pipette_data = self._pipette_store.state.get_pipette_data_by_id(
            pipette_id
        )

        mount = pipette_data.mount
        critical_point = None

        # if the pipette was last used to move to a labware that requires
        # centering, set the critical point to XY_CENTER
        if (
            current_loc is not None and
            current_loc.pipette_id == pipette_id and
            self._labware_store.state.get_labware_has_quirk(
                current_loc.labware_id,
                "centerMultichannelOnWells"
            )
        ):
            critical_point = CriticalPoint.XY_CENTER

        return PipetteLocationData(mount=mount, critical_point=critical_point)

    def get_movement_waypoints(
        self,
        pipette_id: str,
        labware_id: str,
        well_name: str,
        well_location: Optional[WellLocation],
        origin: Point,
        origin_cp: Optional[CriticalPoint],
        max_travel_z: float,
        current_location: Optional[DeckLocation] = None,
    ) -> List[Waypoint]:
        """Get the movement waypoints from an origin to a given location."""
        location = current_location or self.get_current_deck_location()
        center_dest = self._labware_store.state.get_labware_has_quirk(
            labware_id,
            "centerMultichannelOnWells",
        )

        dest = self._geometry_store.state.get_well_position(
            labware_id,
            well_name,
            well_location,
        )
        dest_cp = CriticalPoint.XY_CENTER if center_dest else None

        if (
            location is not None and
            pipette_id == location.pipette_id and
            labware_id == location.labware_id
        ):
            move_type = (
                MoveType.IN_LABWARE_ARC
                if well_name != location.well_name else
                MoveType.DIRECT
            )
            min_travel_z = self._geometry_store.state.get_labware_highest_z(
                labware_id
            )
        else:
            move_type = MoveType.GENERAL_ARC
            min_travel_z = self._geometry_store.state.\
                get_all_labware_highest_z()

        try:
            # TODO(mc, 2021-01-08): inject `get_waypoints` via constructor
            return get_waypoints(
                move_type=move_type,
                origin=origin,
                origin_cp=origin_cp,
                dest=dest,
                dest_cp=dest_cp,
                min_travel_z=min_travel_z,
                max_travel_z=max_travel_z,
                xy_waypoints=[],
            )
        except MotionPlanningError as error:
            raise errors.FailedToPlanMoveError(str(error))


class MotionStore(Substore[MotionState], CommandReactive):
    """Motion state container."""

    _state: MotionState

    def __init__(
        self,
        labware_store: LabwareStore,
        pipette_store: PipetteStore,
        geometry_store: GeometryStore,
    ) -> None:
        """Initialize a MotionStore and its state."""
        self._state = MotionState(
            labware_store=labware_store,
            pipette_store=pipette_store,
            geometry_store=geometry_store,
        )

    def handle_completed_command(
        self,
        command: commands.CompletedCommandType
    ) -> None:
        """Modify state in reaction to a CompletedCommand."""
        if isinstance(
            command.result,
            (
                commands.MoveToWellResult,
                commands.PickUpTipResult,
                commands.DropTipResult,
                commands.AspirateResult,
                commands.DispenseResult,
            ),
        ):
            self._state._current_location = DeckLocation(
                pipette_id=command.request.pipetteId,
                labware_id=command.request.labwareId,
                well_name=command.request.wellName,
            )



"""Protocol engine commands sub-state."""
from typing import Dict, List, Optional, Tuple

from ..commands import CommandType
from .substore import Substore


class CommandState:
    """Command state and getters."""

    _commands_by_id: Dict[str, CommandType]

    def __init__(self) -> None:
        """Initialize a CommandState instance."""
        self._commands_by_id = {}

    def get_command_by_id(self, uid: str) -> Optional[CommandType]:
        """Get a command by its unique identifier."""
        return self._commands_by_id.get(uid)

    def get_all_commands(self) -> List[Tuple[str, CommandType]]:
        """Get a list of all command entries in state."""
        return [entry for entry in self._commands_by_id.items()]


class CommandStore(Substore[CommandState]):
    """Command state container."""

    _state: CommandState

    def __init__(self) -> None:
        """Initialize a CommandStore and its state."""
        self._state = CommandState()

    def handle_command(self, command: CommandType, command_id: str) -> None:
        """Modify state in reaction to any command."""
        self._state._commands_by_id[command_id] = command



"""Movement planning errors."""
from opentrons.types import Point


class MotionPlanningError(Exception):
    """Sub-class for errors from the movement_planning module."""

    def __init__(
        self,
        origin: Point,
        dest: Point,
        clearance: float,
        min_travel_z: float,
        max_travel_z: float,
        message: str
    ) -> None:
        """Initialize an error with properties of the planned motion."""
        super().__init__(message)
        self.origin = origin
        self.dest = dest
        self.clearance = clearance
        self.min_travel_z = min_travel_z
        self.max_travel_z = max_travel_z


class DestinationOutOfBoundsError(MotionPlanningError, ValueError):
    """An error raised when a requested destination is out of bounds."""

    pass


class ArcOutOfBoundsError(MotionPlanningError, ValueError):
    """An error raised when a calculated movement arc is out of bounds."""

    pass



from asyncio import AbstractEventLoop
from typing import Union
from opentrons.broker import Notifications, Broker
from opentrons.hardware_control import ThreadManager, ThreadedAsyncLock
from .session import SessionManager, Session
from .dev_types import Message as SessionMessage
from .calibration import (CalibrationManager, Message as CalibrationMessage)


class MainRouter:
    def __init__(
            self,
            hardware: ThreadManager,
            loop: AbstractEventLoop = None,
            lock: ThreadedAsyncLock = None) -> None:
        topics = [Session.TOPIC, CalibrationManager.TOPIC]
        self._broker = Broker()
        self._notifications: Notifications[
            Union[SessionMessage, CalibrationMessage]] = Notifications(
                topics, self._broker, loop=loop)

        checked_hw = hardware.sync
        self.session_manager = SessionManager(
            hardware=checked_hw,
            loop=loop,
            broker=self._broker,
            lock=lock)
        self.calibration_manager = CalibrationManager(hardware=checked_hw,
                                                      loop=loop,
                                                      broker=self._broker,
                                                      lock=lock)

    @property
    def notifications(self) -> Notifications[
            Union[SessionMessage, CalibrationMessage]]:
        return self._notifications

    @property
    def broker(self) -> Broker:
        return self._broker
